import logging
import fmcapi
import time
import json


def test__devicerecords(fmc):
    logging.info(
        'Test Device.  Though you can "Post" devices I do not have one handy. So '
        "add/remove licenses on Device Objects."
    )

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    all_device_records = fmcapi.DeviceRecords(fmc=fmc)
    all_device_records.get(expanded=True, includeOtherAssociatedPolicies=True)
    # Note: attached RAVPN policies output from 'includeOtherAssociatedPolicies=True'
    # have the INCORRECT UUID. Bug ID: CSCwj27112. This can be worked around for now
    # by subtracting 1 from the UUID in the response. This has ONLY been observed with
    # devicerecords api + 'includeOtherAssociatedPolicies=True' + ravpn policies.

    acp1 = fmcapi.AccessPolicies(fmc=fmc, name=namer)
    acp1.post()
    obj1 = fmcapi.DeviceRecords(fmc=fmc)
    obj1.name = namer
    obj1.acp(name=acp1.name)
    obj1.type = "Device"
    obj1.licensing(action="add", name="MALWARE")
    obj1.licensing(action="add", name="VPN")
    obj1.licensing(action="remove", name="VPN")
    obj1.licensing(action="clear")
    obj1.licensing(action="add", name="BASE")
    obj1.tiering(action="add", name="FTDv5")
    logging.info("Device -->")
    logging.info(obj1.format_data())

    acp1.delete()

    logging.info("Test Device done.\n")



"""Manual NAT Rules Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftdnatpolicies import FTDNatPolicies
from fmcapi.api_objects.object_services.networkaddresses import NetworkAddresses
from fmcapi.api_objects.object_services.networkgroups import NetworkGroups
from fmcapi.api_objects.object_services.portobjectgroups import PortObjectGroups
from fmcapi.api_objects.object_services.protocolportobjects import ProtocolPortObjects
from fmcapi.api_objects.object_services.interfaceobjects import InterfaceObjects
import logging


class ManualNatRules(APIClassTemplate):
    # Host,Network,NetworkGroup objects
    """The ManualNatRules Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "originalSource",
        "originalDestination",
        "translatedSource",
        "translatedDestination",
        "interfaceInTranslatedSource",
        "interfaceInOriginalDestination",
        "natType",
        "interfaceIpv6",
        "fallThrough",
        "dns",
        "routeLookup",
        "noProxyArp",
        "netToNet",
        "sourceInterface",
        "destinationInterface",
        "originalSourcePort",
        "translatedSourcePort",
        "originalDestinationPort",
        "translatedDestinationPort",
        "patOptions",
        "unidirectional",
        "enabled",
        "description",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "section",
    ]
    PREFIX_URL = "/policy/ftdnatpolicies"
    REQUIRED_FOR_POST = ["nat_id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize ManualNatRules object.

        Set self.type to "ManualNatRules" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for ManualNatRules class.")
        self.parse_kwargs(**kwargs)
        self.type = "FTDManualNatRule"
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    @property
    def URL_SUFFIX(self):
        """
        Add the URL suffixes for section, before_auto and after_auto
        NOTE: You must specify these at the time the object is initialized (created) for this feature
        to work correctly. Example:
            This works:
                new_nat_rule = ManualNatRules(fmc=fmc, section="after_auto")

            This does not:
                new_nat_rule = ManualNatRules(fmc=fmc)
                new_nat_rule.section = "after_auto"
        """
        url = "?"

        if "section" in self.__dict__:
            url = f"{url}section={self.section}&"

        return url[:-1]

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for ManualNatRules class.")
        if "translatedSource" in kwargs and "interfaceInTranslatedSource" is True:
            logging.warning(
                "Cannot have both a translatedSource and interfaceInTranslatedSource"
            )
        elif "translatedSource" in kwargs:
            self.translatedSource = kwargs["translatedSource"]
        elif "interfaceInTranslatedSource" in kwargs:
            self.interfaceInTranslatedSource = kwargs["interfaceInTranslatedSource"]

        if "section" in kwargs:
            self.section = kwargs["section"]

    def nat_policy(self, name):
        """
        Associate NAT Policy.

        :param name: (str) Name of NAT Policy.
        :return: None
        """
        logging.debug("In nat_policy() for ManualNatRules class.")
        ftd_nat = FTDNatPolicies(fmc=self.fmc)
        ftd_nat.get(name=name)
        if "id" in ftd_nat.__dict__:
            self.nat_id = ftd_nat.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.nat_id}/manualnatrules{self.URL_SUFFIX}"
            self.nat_added_to_url = True
        else:
            logging.warning(
                f'FTD NAT Policy "{name}" not found.  Cannot set up ManualNatRule for NAT Policy.'
            )

    def original_source(self, name):
        """
        Associate Network to be used as Original Source.

        :param name: (str) Name of Network.
        :return: None
        """
        logging.debug("In original_source() for ManualNatRules class.")
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        networkgroup_json = NetworkGroups(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", []) + networkgroup_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to original_source.'
            )
        else:
            self.originalSource = new_net
            logging.info(f'Adding "{name}" to original_source for this ManualNatRule.')

    def translated_source(self, name):
        """
        Associate Network to be used as Translated Source.

        :param name: (str) Name of Network.
        :return: None
        """
        logging.debug("In translated_source() for ManualNatRules class.")
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        networkgroup_json = NetworkGroups(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", []) + networkgroup_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to translated_source.'
            )
        else:
            self.translatedSource = new_net
            logging.info(
                f'Adding "{name}" to translated_source for this ManualNatRule.'
            )

    def original_destination(self, name):
        """
        Associate Network to be used as Original Destination.

        :param name: (str) Name of Network.
        :return: None
        """
        logging.debug("In original_destination() for ManualNatRules class.")
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        networkgroup_json = NetworkGroups(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", []) + networkgroup_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to original_destination.'
            )
        else:
            self.originalDestination = new_net
            logging.info(
                f'Adding "{name}" to original_destination for this ManualNatRule.'
            )

    def translated_destination(self, name):
        """
        Associate Network to be used as Translated Destination.

        :param name: (str) Name of Network.
        :return: None
        """
        logging.debug("In translated_destination() for ManualNatRules class.")
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        networkgroup_json = NetworkGroups(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", []) + networkgroup_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to translated_destination.'
            )
        else:
            self.translatedDestination = new_net
            logging.info(
                f'Adding "{name}" to translated_destination for this ManualNatRule.'
            )

    def original_source_port(self, name):
        """
        Associate Port to be used as Origin Source port.

        :param name: (str) Name of Port.
        :return: None
        """
        logging.debug("In original_source_port() for ManualNatRules class.")
        ports_json = ProtocolPortObjects(fmc=self.fmc).get()
        portgroup_json = PortObjectGroups(fmc=self.fmc).get()
        items = ports_json.get("items", []) + portgroup_json.get("items", [])
        new_port = None
        for item in items:
            if item["name"] == name:
                new_port = {"id": item["id"], "type": item["type"]}
                break
        if new_port is None:
            logging.warning(
                f'Port "{name}" is not found in FMC.  Cannot add to original_source_port.'
            )
        else:
            self.originalSourcePort = new_port
            logging.info(
                f'Adding "{name}" to original_source_port for this ManualNatRule.'
            )

    def translated_source_port(self, name):
        """
        Associate Port to be used as Translated Source port.

        :param name: (str) Name of Port.
        :return: None
        """
        logging.debug("In translated_source_port() for ManualNatRules class.")
        ports_json = ProtocolPortObjects(fmc=self.fmc).get()
        portgroup_json = PortObjectGroups(fmc=self.fmc).get()
        items = ports_json.get("items", []) + portgroup_json.get("items", [])
        new_port = None
        for item in items:
            if item["name"] == name:
                new_port = {"id": item["id"], "type": item["type"]}
                break
        if new_port is None:
            logging.warning(
                f'Port "{name}" is not found in FMC.  Cannot add to translated_source_port.'
            )
        else:
            self.translatedSourcePort = new_port
            logging.info(
                f'Adding "{name}" to translated_source_port for this ManualNatRule.'
            )

    def original_destination_port(self, name):
        """
        Associate Port to be used as Origin Destination port.

        :param name: (str) Name of Port.
        :return: None
        """
        logging.debug("In original_destination_port() for ManualNatRules class.")
        ports_json = ProtocolPortObjects(fmc=self.fmc).get()
        portgroup_json = PortObjectGroups(fmc=self.fmc).get()
        items = ports_json.get("items", []) + portgroup_json.get("items", [])
        new_port = None
        for item in items:
            if item["name"] == name:
                new_port = {"id": item["id"], "type": item["type"]}
                break
        if new_port is None:
            logging.warning(
                f'Port "{name}" is not found in FMC.  Cannot add to original_destination_port.'
            )
        else:
            self.originalDestinationPort = new_port
            logging.info(
                f'Adding "{name}" to original_destination_port for this ManualNatRule.'
            )

    def translated_destination_port(self, name):
        """
        Associate Port to be used as Destination port.

        :param name: (str) Name of Port.
        :return: None
        """
        logging.debug("In translated_destination_port() for ManualNatRules class.")
        ports_json = ProtocolPortObjects(fmc=self.fmc).get()
        portgroup_json = PortObjectGroups(fmc=self.fmc).get()
        items = ports_json.get("items", []) + portgroup_json.get("items", [])
        new_port = None
        for item in items:
            if item["name"] == name:
                new_port = {"id": item["id"], "type": item["type"]}
                break
        if new_port is None:
            logging.warning(
                f'Port "{name}" is not found in FMC.  Cannot add to translated_destination_port.'
            )
        else:
            self.translatedDestinationPort = new_port
            logging.info(
                f'Adding "{name}" to translated_destination_port for this ManualNatRule.'
            )

    def source_intf(self, name):
        """
        Associate source interface.

        :param name: (str) Name of interface.
        :return: None
        """
        logging.debug("In source_intf() for ManualNatRules class.")
        intf_obj = InterfaceObjects(fmc=self.fmc).get()
        items = intf_obj.get("items", [])
        new_intf = None
        for item in items:
            if item["name"] == name:
                new_intf = {"id": item["id"], "type": item["type"]}
                break
        if new_intf is None:
            logging.warning(
                f'Interface Object "{name}" is not found in FMC.  Cannot add to sourceInterface.'
            )
        else:
            self.sourceInterface = new_intf
            logging.info(f'Interface Object "{name}" added to NAT Policy.')

    def destination_intf(self, name):
        """
        Associate destination interface.

        :param name: (str) Name of interface.
        :return: None
        """
        logging.debug("In destination_intf() for ManualNatRules class.")
        intf_obj = InterfaceObjects(fmc=self.fmc).get()
        items = intf_obj.get("items", [])
        new_intf = None
        for item in items:
            if item["name"] == name:
                new_intf = {"id": item["id"], "type": item["type"]}
                break
        if new_intf is None:
            logging.warning(
                f'Interface Object "{name}" is not found in FMC.  Cannot add to destinationInterface.'
            )
        else:
            self.destinationInterface = new_intf
            logging.info(f'Interface Object "{name}" added to NAT Policy.')

    def identity_nat(self, name):
        """
        Associate an Identity Network.

        :param name: (str) Name of Network.
        :return: None
        """
        logging.debug("In identity_nat() for ManualNatRules class.")
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        networkgroup_json = NetworkGroups(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", []) + networkgroup_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to this ManualNatRules.'
            )
        else:
            self.natType = "STATIC"
            self.originalSource = new_net
            self.translatedSource = new_net
            logging.info(f'Adding "{name}" to ManualNatRules.')

    def patPool(self, name, options={}):
        """
        Associate a PAT Pool.

        :param name: (str) Name of PAT Pool.
        :param options: (dict) key/value of options.
        :return: None
        """
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        networkgroup_json = NetworkGroups(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", []) + networkgroup_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"name": item["name"], "id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to patPool.'
            )
        else:
            self.natType = "DYNAMIC"
            self.unidirectional = True
            self.patOptions = {"patPoolAddress": new_net}
            self.patOptions["interfacePat"] = (
                options["interfacePat"] if "interfacePat" in options.keys() else False
            )
            self.patOptions["includeReserve"] = (
                options["includeReserve"]
                if "includeReserve" in options.keys()
                else False
            )
            self.patOptions["roundRobin"] = (
                options["roundRobin"] if "roundRobin" in options.keys() else True
            )
            self.patOptions["extendedPat"] = (
                options["extendedPat"] if "extendedPat" in options.keys() else False
            )
            self.patOptions["flatPortRange"] = (
                options["flatPortRange"] if "flatPortRange" in options.keys() else False
            )
            logging.info(f'Adding "{name}" to patPool for this ManualNatRule.')



"""Port Object Groups Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ports import Ports
import logging


class PortObjectGroups(APIClassTemplate):
    """The PortObjectGroups Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type", "objects", "literals", "description"]
    VALID_GET_FILTERS = [
        "unusedOnly",
        "nameOrValue",
    ]  # unusedOnly:Bool, nameOrValue:String
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/portobjectgroups"

    # Technically you can have objects OR literals but I'm not set up for "OR" logic, yet.
    REQUIRED_FOR_POST = ["name", "objects"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize PortObjectGroups object.

        Set self.type to "PortObjectGroup" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for PortObjectGroups class.")
        self.parse_kwargs(**kwargs)
        self.type = "PortObjectGroup"

    def named_ports(self, action, name=""):
        """
        Associate Named Ports.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of port.
        """
        logging.debug("In named_ports() for PortObjectGroups class.")
        if action == "add":
            port1 = Ports(fmc=self.fmc)
            response = port1.get()
            if "items" in response:
                new_port = None
                for item in response["items"]:
                    if item["name"] == name:
                        new_port = {
                            "name": item["name"],
                            "id": item["id"],
                            "type": item["type"],
                        }
                        break
                if new_port is None:
                    logging.warning(
                        f'Port "{name}" is not found in FMC.  Cannot add to PortObjectGroups.'
                    )
                else:
                    if "objects" in self.__dict__:
                        duplicate = False
                        for obj in self.objects:
                            if obj["name"] == new_port["name"]:
                                duplicate = True
                                break
                        if not duplicate:
                            self.objects.append(new_port)
                            logging.info(f'Adding "{name}" to PortObjectGroups.')
                    else:
                        self.objects = [new_port]
                        logging.info(f'Adding "{name}" to PortObjectGroups.')
        elif action == "remove":
            if "objects" in self.__dict__:
                objects_list = []
                for obj in self.objects:
                    if obj["name"] != name:
                        objects_list.append(obj)
                self.objects = objects_list
                logging.info(f'Removed "{name}" from PortObjectGroups.')
            else:
                logging.info(
                    "This PortObjectGroups has no named_ports.  Nothing to remove."
                )
        elif action == "clear":
            if "objects" in self.__dict__:
                del self.objects
                logging.info("All named_ports removed from this PortObjectGroups.")



import logging
import fmcapi


def test__deployment_requests(fmc):
    logging.info("Testing DeploymentRequests() class.")
    tmp = fmcapi.DeploymentRequests(fmc=fmc)
    logging.info(tmp.post())
    logging.info("Testing DeploymentRequests() class done.\n")



import logging
import fmcapi
import time


def test__port_object_group(fmc):
    logging.info("Testing PortObjectGroup class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj10 = fmcapi.ProtocolPortObjects(
        fmc=fmc, name="_porttcp1", port="8443", protocol="TCP"
    )
    obj10.post()
    obj11 = fmcapi.ProtocolPortObjects(
        fmc=fmc, name="_portudp1", port="161", protocol="UDP"
    )
    obj11.post()
    obj12 = fmcapi.ProtocolPortObjects(
        fmc=fmc, name="_portrangetcp1", port="0-1023", protocol="TCP"
    )
    obj12.post()
    time.sleep(1)
    obj1 = fmcapi.PortObjectGroups(fmc=fmc, name=namer)
    obj1.named_ports(action="add", name=obj10.name)
    obj1.named_ports(action="add", name=obj10.name)
    obj1.named_ports(action="remove", name=obj10.name)
    obj1.named_ports(action="clear")
    obj1.named_ports(action="add", name=obj11.name)
    obj1.named_ports(action="add", name=obj12.name)
    obj1.named_ports(action="remove", name=obj11.name)
    obj1.post()
    time.sleep(1)
    del obj1

    obj1 = fmcapi.PortObjectGroups(fmc=fmc, name=namer)
    obj1.get()
    obj1.named_ports(action="add", name="HTTP")
    obj1.named_ports(action="clear")
    obj1.named_ports(action="add", name="HTTP")
    obj1.named_ports(action="remove", name="HTTP")
    obj1.named_ports(action="add", name="HTTP")
    obj1.named_ports(action="add", name="HTTPS")
    obj1.put()

    time.sleep(1)
    obj1.delete()
    obj10.delete()
    obj11.delete()
    obj12.delete()

    logging.info("Testing PortObjectGroup class done.\n")



import logging
import fmcapi


def test__backup(fmc):
    logging.info("Test Backup.  Get, delete Backups.")

    obj1 = fmcapi.Backup(fmc=fmc)
    obj1.get()
    del(obj1)

    # Requires existing backup and there does not seem to be a way to initiate an FMC backup from the API
    # There is a way to initiate a device backup with /backup/operational/devicebackup

    # obj1 = fmcapi.Backup(fmc=fmc)
    # obj1.delete(targetId='manager', backupVersion=1694394505)

    logging.info("Test Backup done.\n")



"""Server Version Class."""

import logging


class ServerVersion(object):  # Can't import APIClassTemplate due to dependency loop.
    """
    Get the FMC's version information.

    Set instance variables for each version info returned as well as return the whole response text.
    :return: requests' response.text
    """

    URL_SUFFIX = "/info/serverversion"

    def __init__(self, fmc):
        """
        Initialize ServerVersion object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        logging.debug("In __init__() for ServerVersion class.")

        self.fmc = fmc
        self.URL = f"{self.fmc.platform_url}{self.URL_SUFFIX}"
        self.vdbVersion = None
        self.sruVersion = None
        self.serverVersion = None
        self.geoVersion = None

    def get(self):
        """
        Send GET to FMC.

        :return: requests response
        """
        logging.debug("GET method for API for ServerVersion.")
        response = self.fmc.send_to_api(method="get", url=self.URL)
        if "items" in response:
            logging.info(
                "Populating vdbVersion, sruVersion, serverVersion, and geoVersion FMC instance variables."
            )
            self.vdbVersion = response["items"][0]["vdbVersion"]
            self.sruVersion = response["items"][0]["sruVersion"]
            self.serverVersion = response["items"][0]["serverVersion"]
            self.geoVersion = response["items"][0]["geoVersion"]
        return response

    def post(self):
        """POST method for API for ServerVersion not supported."""
        logging.info("POST method for API for ServerVersion not supported.")
        pass

    def put(self):
        """PUT method for API for ServerVersion not supported."""
        logging.info("PUT method for API for ServerVersion not supported.")
        pass

    def delete(self):
        """DELETE method for API for ServerVersion not supported."""
        logging.info("DELETE method for API for ServerVersion not supported.")
        pass



"""Moving the fmc.deploymentrequests to an actual api_object."""

import logging
from .deployabledevices import DeployableDevices
import datetime


class DeploymentRequests(
    object
):  # Can't import APIClassTemplate due to dependency loop.
    """
    Iterate through the list of devices needing deployed and submit a request to the FMC to deploy changes to them.

    :return:
    """

    URL_SUFFIX = "/deployment/deploymentrequests"

    def __init__(self, fmc):
        """
        Initialize DeploymentRequests object.

        :param fmc (object): FMC object
        :return: None
        """
        logging.debug("In __init__ for DeploymentRequests() class.")

        self.fmc = fmc
        self.URL = f"{self.fmc.configuration_url}{self.URL_SUFFIX}"
        self.uuids = None

    def get(self):
        """GET method for API for DeploymentRequests not supported."""
        logging.info("GET method for API for DeploymentRequests not supported.")

    def post(self):
        """
        Submit list of devices to FMC that need config changes pushed to them.

        :return: (list) List of devices.
        """
        logging.debug("In post() method for DeploymentRequests() class.")
        devices = DeployableDevices(fmc=self.fmc)
        self.uuids = devices.get()
        if not self.uuids:
            logging.info("No devices need deployed.")
            return
        json_data = {
            "type": "DeploymentRequest",
            "forceDeploy": True,
            "ignoreWarning": True,
            "version": str(int(1000000 * datetime.datetime.utcnow().timestamp())),
            "deviceList": [],
        }
        for device in self.uuids:
            logging.info(f"Adding device {device} to deployment queue.")
            json_data["deviceList"].append(device["device"]["id"])
            # From the list of deployable devices get the version value that is smallest.
            if int(json_data["version"]) > int(device["version"]):
                logging.info(f"Updating version to {device['version']}")
                json_data["version"] = device["version"]
        logging.info("Deploying changes to devices.")
        response = self.fmc.send_to_api(
            method="post", url=self.URL, json_data=json_data
        )
        return response["deviceList"]

    def put(self):
        """PUT method for API for DeploymentRequests not supported."""
        logging.info("PUT method for API for DeploymentRequests not supported.")
        pass

    def delete(self):
        """DELETE method for API for DeploymentRequests not supported."""
        logging.info("DELETE method for API for DeploymentRequests not supported.")
        pass



import logging
import fmcapi


def test__filepolicies(fmc):
    logging.info("Testing FilePolicies class.")

    obj1 = fmcapi.FilePolicies(fmc=fmc)
    logging.info("All FilePolicies -- >")
    result = obj1.get()
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")

    del obj1

    logging.info("Testing FilePolicies class done.\n")



"""Upgrades Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.device_services.devicerecords import DeviceRecords
from .upgradepackages import UpgradePackages
import logging


class Upgrades(APIClassTemplate):
    """
    The Upgrades Object in the FMC.

    NOTE:  This should be called UpgradePackage but that collides with a Deprecated name for UpgradePackages.
    We can rename this after we remove that deprecation... which will be a while from now.
    """

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "upgradePackage",
        "targets",
        "pushUpgradeFileOnly",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/updates/upgrades"

    def __init__(self, fmc, **kwargs):
        """
        Initialize Upgrades object.

        Set self.type to "Upgrade", parse the kwargs, and set up the self.URL.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Upgrades class.")
        self.type = "Upgrade"
        self.URL = f"{self.fmc.platform_url}{self.URL_SUFFIX}"
        self.parse_kwargs(**kwargs)

    def upgrade_package(self, package_name):
        """
        Upgrade named package.

        :param package_name: (str) Name of package to upgrade
        :return: None
        """
        logging.debug("In upgrade_package() for Upgrades class.")
        package1 = UpgradePackages(fmc=self.fmc)
        package1.get(name=package_name)
        if "id" in package1.__dict__:
            self.upgradePackage = {"id": package1.id, "type": package1.type}
        else:
            logging.warning(
                f'UpgradePackage "{package_name}" not found.  Cannot add package to Upgrades.'
            )

    def devices(self, devices):
        """
        List of devices.

        :param devices: (list) List of device names.
        :return: None
        """
        logging.debug("In devices() for Upgrades class.")
        for device in devices:
            device1 = DeviceRecords(fmc=self.fmc)
            device1.get(name=device)
            if "id" in device1.__dict__ and "targets" in self.__dict__:
                self.targets.append(
                    {"id": device1.id, "type": device1.type, "name": device1.name}
                )
            elif "id" in device1.__dict__:
                self.targets = [
                    {"id": device1.id, "type": device1.type, "name": device1.name}
                ]
            else:
                logging.warning(
                    f'Device "{device}" not found.  Cannot prepare devices for Upgrades.'
                )

    def get(self):
        """GET method for API for Upgrades not supported."""
        logging.info("GET method for API for Upgrades not supported.")
        pass

    def post(self, **kwargs):
        """Run POST call on FMC."""
        # returns a task status object
        logging.debug("In post() for Upgrades class.")
        self.fmc.autodeploy = False
        return super().post(**kwargs)

    def put(self):
        """PUT method for API for Upgrades not supported."""
        logging.info("PUT method for API for Upgrades not supported.")
        pass

    def delete(self):
        """DELETE method for API for Upgrades not supported."""
        logging.info("DELETE method for API for Upgrades not supported.")
        pass



import logging
import fmcapi
import time


def test__ip_network(fmc):
    logging.info("Test IPNetwork.  Post, get, put, delete Network Objects.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj1 = fmcapi.Networks(fmc=fmc)
    obj1.name = namer
    obj1.value = "8.8.8.0/24"
    obj1.post()
    time.sleep(1)
    del obj1

    obj1 = fmcapi.Networks(fmc=fmc, name=namer)
    obj1.get()
    obj1.value = "9.9.9.0/24"
    obj1.put()
    time.sleep(1)
    obj1.delete()

    logging.info("Test IPNetwork done.\n")



import logging
import fmcapi


def test__cert_enrollment(fmc):
    logging.info("Testing CertEnrollment class. Requires a CertEnrollment")

    obj1 = fmcapi.CertEnrollments(fmc=fmc)
    logging.info("All CertEnrollments -- >")
    result = obj1.get()
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")
    del obj1

    logging.info("Testing CertEnrollment class done.\n")



"""IKE v2 IPSec Proposals Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class IKEv2IpsecProposals(APIClassTemplate):
    """The IKEv2IpsecProposals Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "encryptionAlgorithms",
        "integrityAlgorithms",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/ikev2ipsecproposals"
    REQUIRED_FOR_POST = ["name", "encryptionAlgorithms", "integrityAlgorithms"]
    VALID_FOR_ENCRYPTION = [
        "DES",
        "3DES",
        "AES",
        "AES-192",
        "AES-256",
        "NULL",
        "AES-GCM",
        "AES-GCM-192",
        "AES-GCM-256",
        "AES-GMAC",
        "AES-GMAC-192",
        "AES-GMAC-256",
    ]
    VALID_FOR_HASH = ["NULL", "MD5", "SHA-1", "SHA-256", "SHA-384", "SHA-512"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""
    FIRST_SUPPORTED_FMC_VERSION = "6.3"

    def __init__(self, fmc, **kwargs):
        """
        Initialize IKEv2IpsecProposals object.

        Set self.type to "IKEv2IpsecProposal" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for IKEv2IpsecProposals class.")
        self.parse_kwargs(**kwargs)
        self.type = "IKEv2IPsecProposal"

    def encryption(self, action, algorithms=[]):
        """
        Associate encryption.

        :param action: (str) 'add', 'remove', or 'clear'
        :param algorithms: (list) List of algorithms.
        :return: None
        """
        logging.debug("In encryption() for IKEv2IpsecProposals class.")
        if action == "add":
            for algorithm in algorithms:
                if "encryptionAlgorithms" in self.__dict__:
                    if algorithm in self.encryptionAlgorithms:
                        logging.warning(
                            f'encryptionAlgorithms "{algorithm}" already exists.'
                        )
                    elif algorithm in self.VALID_FOR_ENCRYPTION:
                        self.encryptionAlgorithms.append(algorithm)
                    else:
                        logging.warning(
                            f'encryptionAlgorithms "{algorithm}" not a valid type.'
                        )
                else:
                    self.encryptionAlgorithms = [algorithm]
        elif action == "remove":
            if "encryptionAlgorithms" in self.__dict__:
                for algorithm in algorithms:
                    self.encryptionAlgorithms = list(
                        filter(lambda i: i != algorithm, self.encryptionAlgorithms)
                    )
            else:
                logging.warning(
                    "IKEv2IpsecProposals has no members.  Cannot remove encryptionAlgorithms."
                )
        elif action == "clear":
            if "encryptionAlgorithms" in self.__dict__:
                del self.encryptionAlgorithms
                logging.info(
                    "All encryptionAlgorithms removed from this IKEv2IpsecProposals object."
                )

    def hash(self, action, algorithms=[]):
        """
        Associate hash.

        :param action: (str) 'add', 'remove', or 'clear'
        :param algorithms: (list) List of algorithms.
        :return: None
        """
        logging.debug("In hash() for IKEv2IpsecProposals class.")
        if action == "add":
            for algorithm in algorithms:
                if "integrityAlgorithms" in self.__dict__:
                    if algorithm in self.integrityAlgorithms:
                        logging.warning(
                            f'integrityAlgorithms "{algorithm}" already exists.'
                        )
                    elif algorithm in self.VALID_FOR_HASH:
                        self.integrityAlgorithms.append(algorithm)
                    else:
                        logging.warning(
                            f'integrityAlgorithms "{algorithm}" not a valid type.'
                        )
                else:
                    self.integrityAlgorithms = [algorithm]
        elif action == "remove":
            if "integrityAlgorithms" in self.__dict__:
                for algorithm in algorithms:
                    self.integrityAlgorithms = list(
                        filter(lambda i: i != algorithm, self.integrityAlgorithms)
                    )
            else:
                logging.warning(
                    "IKEv2IpsecProposals has no members.  Cannot remove integrityAlgorithms."
                )
        elif action == "clear":
            if "integrityAlgorithms" in self.__dict__:
                del self.integrityAlgorithms
                logging.info(
                    "All integrityAlgorithms removed from this IKEv2IpsecProposals object."
                )



from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class TerminateRAVPNSessions(APIClassTemplate):
    """The TerminateRAVPNSessions Object in the FMC."""

    FIRST_SUPPORTED_FMC_VERSION = "7.3"
    VALID_JSON_DATA = [
        "id",
        "name",
        "description",
        "usernames",
        "type",
        "deviceId",
        "terminateBy",
        "sessionIds",
        "version",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    REQUIRED_FOR_POST = ["terminateBy"]

    URL_SUFFIX = "/health/ravpnsessions/operational/terminateravpnsessions"

    def __init__(self, fmc, **kwargs):
        """
        Initialize TerminateRAVPNSessions object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: requests response
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for TerminateRAVPNSessions class.")
        self.parse_kwargs(**kwargs)

    def get(self, **kwargs):
        """GET method for API for TerminateRAVPNSessions not supported."""
        logging.info("GET method for API for TerminateRAVPNSessions not supported.")
        pass

    def delete(self, **kwargs):
        """DELETE method for API for TerminateRAVPNSessions not supported."""
        logging.info("DELETE method for API for TerminateRAVPNSessions not supported.")
        pass

    def put(self):
        """PUT method for API for TerminateRAVPNSessions not supported."""
        logging.info("PUT method for API for TerminateRAVPNSessions not supported.")
        pass



from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class Backup(APIClassTemplate):
    """The Backup Object in the FMC."""

    FIRST_SUPPORTED_FMC_VERSION = "7.3"
    VALID_JSON_DATA = []
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "targetId",
        "backupVersion",
    ]
    REQUIRED_FOR_DELETE = [
        "targetId",
        "backupVersion",
    ]

    URL_SUFFIX = "/backup/files"

    def __init__(self, fmc, **kwargs):
        """
        Initialize Backup object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: requests response
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Backup class.")
        self.parse_kwargs(**kwargs)

    def get(self, **kwargs):
        """
        Prepare to send GET call to FMC Backup Files API.

        If no self.targetId exists then return a full listing of all
        backups. Self.targetId must be set to 'manager' to target an FMC backup, otherwise,
        set self.targetId to device UUID to target backups for a particular FTD device.
        Set self.backupVersion(optional) with self.targetId to a specific backup version ID
        otherwise the latest backup version of self.targetId will be used.
        Set "expanded=true" results for specific object
        to gather additional detail.

        :return: requests response
        """
        response = super().get(**kwargs)
        return response

    def delete(self, **kwargs):
        """
        Prepare to send DELETE call to FMC Backup Files API.

        Self.targetId is required.
        Set self.targetId must be set to 'manager' to target an FMC backup, otherwise,
        set self.targetId to device UUID to target backups for a particular FTD device.
        Set self.backupVersion is technically optional, but is being enforced as required for extra safety.
        Otherwise, the latest backup could be deleted unintentionally.
        Set "expanded=true" results for specific object
        to gather additional detail.

        :return: requests response
        """
        response = super().delete(**kwargs)
        return response

    def post(self):
        """POST method for API for Backup not supported."""
        logging.info("POST method for API for Backup not supported.")
        pass

    def put(self):
        """PUT method for API for Backup not supported."""
        logging.info("PUT method for API for Backup not supported.")
        pass



import logging
import fmcapi


def test__object(fmc):
    logging.info("Test Object.  Get search results.")

    obj1 = fmcapi.Object(fmc=fmc, filter='0.0.0.0')
    result = obj1.get()
    logging.info(f"Total items: {len(result['items'])}")

    logging.info("Test Object done.\n")



"""Upgrade Packages Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class UpgradePackages(APIClassTemplate):
    """The UpgradePackages Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/updates/upgradepackages"

    def __init__(self, fmc, **kwargs):
        """
        Initialize UpgradePackages object.

        Set self.type to "UpgradePackage", parse the kwargs, and set up the self.URL.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for UpgradePackages class.")
        self.type = "UpgradePackage"
        self.URL = f"{self.fmc.platform_url}{self.URL_SUFFIX}"
        self.parse_kwargs(**kwargs)

    def post(self):
        """POST method for API for UpgradePackages not supported."""
        logging.info("POST method for API for UpgradePackages not supported.")
        pass

    def put(self):
        """PUT method for API for UpgradePackages not supported."""
        logging.info("PUT method for API for UpgradePackages not supported.")
        pass



import logging
import fmcapi
import time


def test__interface_group(fmc):
    logging.info("Test InterfaceGroup.  Post, get, put, delete InterfaceGroup Objects.")

    obj1 = fmcapi.InterfaceGroups(fmc=fmc)
    obj1.name = "_ig_outside_all"
    obj1.interfaceMode = "ROUTED"
    logging.info("InterfaceGroup POST-->")
    logging.info(obj1.format_data())
    logging.info("\n")
    obj1.post()
    time.sleep(1)
    del obj1

    obj1 = fmcapi.InterfaceGroups(fmc=fmc, name="_ig_outside_all")
    obj1.get()
    obj1.p_interface(
        device_name="device_name",
        action="add",
        names=["GigabitEthernet0/0", "GigabitEthernet0/1", "GigabitEthernet0/2"],
    )
    logging.info("InterfaceGroup PUT-->")
    logging.info(obj1.format_data())
    logging.info("\n")
    obj1.put()
    time.sleep(1)
    del obj1

    obj1 = fmcapi.InterfaceGroups(fmc=fmc, name="_ig_outside_all")
    obj1.get()
    obj1.p_interface(
        device_name="device_name", action="remove", names=["GigabitEthernet0/1"]
    )
    logging.info("InterfaceGroup PUT-->")
    logging.info(obj1.format_data())
    logging.info("\n")
    obj1.put()
    time.sleep(1)
    del obj1

    obj1 = fmcapi.InterfaceGroups(fmc=fmc, name="_ig_outside_all")
    obj1.get()
    obj1.p_interface(action="clear-all")
    obj1.put()
    logging.info("InterfaceGroup DELETE-->")
    logging.info(obj1.format_data())
    logging.info("\n")
    obj1.delete()
    del obj1

    logging.info("Test InterfaceGroup done.\n")



"""SLA Monitors Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .securityzones import SecurityZones
import logging


class SLAMonitors(APIClassTemplate):
    """The SLAMonitors Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "timeout",
        "threshold",
        "frequency",
        "slaId",
        "dataSize",
        "tos",
        "noOfPackets",
        "monitorAddress",
        "interfaceObjects",
        "description",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/slamonitors"
    REQUIRED_FOR_POST = ["name", "slaId", "monitorAddress", "interfaceObjects", "type"]
    REQUIRED_FOR_PUT = ["id", "type"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize SLAMonitors object.

        Set self.type to "SLAMonitor" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for SLAMonitors class.")
        self.parse_kwargs(**kwargs)
        self.type = "SLAMonitor"

    def interfaces(self, names):
        """
        Associate interfaces (AKA Security Zones).

        :param names: (list) List of Security Zone names.
        """
        logging.debug("In interfaces() for SLAMonitors class.")
        zones = []
        for name in names:
            # Supports passing list of str
            sz = SecurityZones(fmc=self.fmc)
            sz.get(name=name)
            if "id" in sz.__dict__:
                zones.append({"name": sz.name, "id": sz.id, "type": sz.type})
            else:
                logging.warning(
                    f'Security Zone, "{name}", not found.  Cannot add to SLAMonitors.'
                )
        if len(zones) != 0:
            # Make sure we found at least one zone
            self.interfaceObjects = zones
        else:
            logging.warning(
                f'No valid Security Zones found: "{names}".  Cannot add to SLAMonitosr.'
            )



import logging
import time
from fmcapi import (
    Hosts,
    Networks,
    Ranges,
    FQDNS,
    NetworkGroups,
    PreFilterPolicies,
    PreFilterRules,
    SecurityZones,
    PortObjectGroups,
    ProtocolPortObjects,
    VlanGroupTags,
    VlanTags,
)


def test__prefiler_rule(fmc):
    logging.info("Testing prefilter rules on FMC")
    logging.info("Creating objects for testing on FMC")

    namer = f"fmcapi_test_{str(int(time.time()))}"
    ip_host_1 = Hosts(fmc=fmc, name=f"test_host_1_{namer}", value="7.7.7.7")
    ip_host_1.post()
    ip_net_1 = Networks(fmc=fmc, name=f"test_net_1_{namer}", value="10.0.0.0/8")
    ip_net_1.post()
    ip_range_1 = Ranges(
        fmc=fmc, name=f"test_range_1_{namer}", value="10.1.1.1-10.1.1.10"
    )
    ip_range_1.post()
    fqdn_1 = FQDNS(fmc=fmc, name=f"test_fqdn_1_{namer}", value="www.cisco.com")
    fqdn_1.post()
    net_group_1 = NetworkGroups(fmc=fmc, name=f"net_group_1_{namer}")
    net_group_1.named_networks(action="add", name=f"test_net_1_{namer}")
    net_group_1.post()
    sec_zone_1 = SecurityZones(
        fmc=fmc, name=f"test_zone_1_{namer}", interfaceMode="ROUTED"
    )
    sec_zone_1.post()
    sec_zone_2 = SecurityZones(
        fmc=fmc, name=f"test_zone_2_{namer}", interfaceMode="ROUTED"
    )
    sec_zone_2.post()
    port_1 = ProtocolPortObjects(
        fmc=fmc, name=f"test_port_1_{namer}", port="8443", protocol="TCP"
    )
    port_1.post()
    port_2 = ProtocolPortObjects(
        fmc=fmc, name=f"test_port_2_{namer}", port="161", protocol="UDP"
    )
    port_2.post()
    port_3 = ProtocolPortObjects(
        fmc=fmc, name=f"test_port_3_{namer}", port="0-1023", protocol="TCP"
    )
    port_3.post()
    time.sleep(1)
    port_group_1 = PortObjectGroups(fmc=fmc, name=f"port_group_1_{namer}")
    port_group_1.named_ports(action="add", name=port_1.name)
    port_group_1.named_ports(action="add", name=port_2.name)
    port_group_1.named_ports(action="add", name=port_3.name)
    port_group_1.post()

    vlan_tag_1 = VlanTags(fmc=fmc, name=f"vlan_tag_1_{namer}")
    vlan_tag_1.vlans(start_vlan="1", end_vlan="9")
    vlan_tag_1.post()
    vlan_tag_2 = VlanTags(fmc=fmc, name=f"vlan_tag_2_{namer}")
    vlan_tag_2.vlans(start_vlan="10", end_vlan="19")
    vlan_tag_2.post()
    vlan_tag_3 = VlanTags(fmc=fmc, name=f"vlan_tag_3_{namer}")
    vlan_tag_3.vlans(start_vlan="20", end_vlan="29")
    vlan_tag_3.post()
    vlan_group_1 = VlanGroupTags(fmc=fmc, name=f"vlan_group_1_{namer}")
    vlan_group_1.named_vlantags(action="add", name=f"vlan_tag_3_{namer}")
    vlan_group_1.post()

    logging.info(f'Creating test prefiler "{namer}"')
    prefilter = PreFilterPolicies(fmc=fmc, name=namer)
    prefilter.post()
    time.sleep(1)

    logging.info(f'Creating test prefiler "{namer}"')
    prefilter = PreFilterPolicies(fmc=fmc, name=namer)
    prefilter.post()
    time.sleep(1)

    logging.info(f'Creating test prefiler rule "test_1"')
    prefilter_rule_1 = PreFilterRules(fmc=fmc, prefilter_name=namer)
    prefilter_rule_1.name = "test_1"
    prefilter_rule_1.enabled = True
    prefilter_rule_1.sendEventsToFMC = True
    prefilter_rule_1.logBegin = True
    prefilter_rule_1.logEnd = True
    prefilter_rule_1.enableSyslog = True
    prefilter_rule_1.source_interface(action="add", name=f"test_zone_1_{namer}")
    prefilter_rule_1.destination_interface(action="add", name=f"test_zone_2_{namer}")
    prefilter_rule_1.source_network(action="add", literal="10.1.1.1")
    prefilter_rule_1.destination_network(action="add", name=f"test_host_1_{namer}")
    prefilter_rule_1.destination_network(action="add", name=f"test_net_1_{namer}")
    prefilter_rule_1.destination_network(action="add", name=f"test_range_1_{namer}")
    prefilter_rule_1.destination_network(action="add", name=f"test_fqdn_1_{namer}")
    prefilter_rule_1.destination_network(action="remove", name=f"test_fqdn_1_{namer}")
    prefilter_rule_1.new_comments(action="add", value="comment-1")
    prefilter_rule_1.new_comments(action="add", value="comment-2")
    prefilter_rule_1.new_comments(action="remove", value="comment-2")
    prefilter_rule_1.post()
    time.sleep(1)

    logging.info(f'Creating test prefiler rule "test_2"')
    prefilter_rule_2 = PreFilterRules(fmc=fmc, prefilter_name=namer)
    prefilter_rule_2.name = "test_2"
    prefilter_rule_2.enabled = True
    prefilter_rule_2.action = "FASTPATH"
    prefilter_rule_2.source_port(action="add", literal={"protocol": "6", "port": "22"})
    prefilter_rule_2.source_port(action="add", literal={"protocol": "6", "port": "443"})
    prefilter_rule_2.source_port(action="add", literal={"protocol": "17", "port": "53"})
    prefilter_rule_2.destination_port(action="add", name=f"test_port_1_{namer}")
    prefilter_rule_2.destination_port(action="add", name=f"test_port_2_{namer}")
    prefilter_rule_2.destination_port(action="add", name=f"test_port_3_{namer}")
    prefilter_rule_2.destination_port(action="remove", name=f"test_port_3{namer}")
    prefilter_rule_2.destination_port(action="add", name=f"port_group_1_{namer}")
    prefilter_rule_2.vlan_tags(action="add", literal="100-110")
    prefilter_rule_2.vlan_tags(action="add", name=f"vlan_tag_1_{namer}")
    prefilter_rule_2.vlan_tags(action="add", name=f"vlan_tag_2_{namer}")
    prefilter_rule_2.vlan_tags(action="add", name=f"vlan_tag_3_{namer}")
    prefilter_rule_2.vlan_tags(action="remove", name=f"vlan_tag_2_{namer}")
    prefilter_rule_2.vlan_tags(action="add", name=f"vlan_group_1_{namer}")
    prefilter_rule_2.post()
    time.sleep(1)

    logging.info(f'Creating test tunnel rule "test_3"')
    prefilter_rule_3 = PreFilterRules(fmc=fmc, prefilter_name=namer)
    prefilter_rule_3.name = "test_3"
    prefilter_rule_3.enabled = True
    prefilter_rule_3.action = "FASTPATH"
    prefilter_rule_3.rule_type("TUNNEL")
    prefilter_rule_3.bidirectional = True
    prefilter_rule_3.encapsulation_port(action="clear")
    prefilter_rule_3.encapsulation_port(action="add", value="GRE")
    prefilter_rule_3.encapsulation_port(action="add", value="IP_IN_IP")
    prefilter_rule_3.encapsulation_port(action="add", value="TEREDO")
    prefilter_rule_3.encapsulation_port(action="remove", value="TEREDO")
    prefilter_rule_3.post()
    time.sleep(1)

    prefilter.delete()
    ip_host_1.delete()
    ip_net_1.delete()
    ip_range_1.delete()
    fqdn_1.delete()
    net_group_1.delete()
    sec_zone_1.delete()
    sec_zone_2.delete()
    port_group_1.delete()
    port_1.delete()
    port_2.delete()
    port_3.delete()
    vlan_group_1.delete()
    vlan_tag_1.delete()
    vlan_tag_2.delete()
    vlan_tag_3.delete()



"""ACP Inheritance Settings Class"""
from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .accesspolicies import AccessPolicies
import logging


class InheritanceSettings(APIClassTemplate):
    """The InheritanceSettings Object in the FMC."""

    VALID_JSON_DATA = []
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "id",
        "acp_id",
        "acp_name",
        "device_id",
        "device_name",
        "base_policy_id",
    ]
    PREFIX_URL = "/policy/accesspolicies"
    REQUIRED_FOR_PUT = ["acp_id", "id", "base_policy_id"]
    REQUIRED_FOR_GET = ["acp_id"]
    FIRST_SUPPORTED_FMC_VERSION = "6.5"

    def __init__(self, fmc, **kwargs):
        """
        Initialize InheritanceSettings object.

        Set self.type to "AccessPolicyInheritanceSettings", parse the kwargs, and set up the self.URL.

        Note: The InheritanceSettings API is a bit of a weird API. The construction of the API URI needs two IDs.
        According to the API documentation you need the ContainerID and the objectId where the later is defined as
        "Unique identifier of the Access Policy Inheritance Setting". These two values are actually the same value,
        specifically the ID of the ACP for which you are changing the inheritance.

        The third ID that is required for this API, which we have called the base_policy_id to align with the API
        documentation, is the ID of ACP that you want to be the parent.

        An example of use would be:

        fmc_integrate = InheritanceSettings(
            fmc=fmc_session,
            acp_id=child_policy["id"],
            id=child_policy["id"],
            base_policy_id=parent_policy["id"],
        )
        fmc_integrate.put()

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        logging.debug("In __init__() for InheritanceSettings class.")
        super().__init__(fmc, **kwargs)
        self.parse_kwargs(**kwargs)
        self.type = "AccessPolicyInheritanceSettings"
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for InheritanceSettings class.")
        if "acp_id" in kwargs:
            self.acp(acp_id=kwargs["acp_id"])
        if "acp_name" in kwargs:
            self.acp(name=kwargs["acp_name"])
        if "device_id" in kwargs:
            self.device(id=kwargs["device_id"])
        if "device_name" in kwargs:
            self.device(name=kwargs["device_name"])

    def acp(self, name="", acp_id=""):
        """
        Associate an AccessPolicies object with this InheritanceSettings object.

        :param name: (str)  Name of ACP.
        :param id: (str) ID of ACP.
        :return: None
        """
        # either name or id of the ACP should be given
        logging.debug("In acp() for InheritanceSettings class.")
        if acp_id != "":
            self.acp_id = acp_id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.acp_id}/inheritancesettings"
            self.acp_added_to_url = True
        elif name != "":
            acp1 = AccessPolicies(fmc=self.fmc)
            acp1.get(name=name)
            if "id" in acp1.__dict__:
                self.acp_id = acp1.id
                self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.acp_id}/inheritancesettings"
                self.acp_added_to_url = True
            else:
                logging.warning(
                    f'Access Control Policy "{name}" not found.  Cannot configure acp for InheritanceSettings.'
                )
        else:
            logging.error("No accessPolicy name or id was provided.")

    def format_data(self, filter_query=""):
        """
        Gather all the data in preparation for sending to API in JSON format.

        :param filter_query: (str) 'all' or 'kwargs'
        :return: (dict) json_data
        """
        logging.debug("In format_data() for InheritanceSettings class.")
        return {
            "type": "AccessPolicyInheritanceSetting",
            "id": self.acp_id,
            "basePolicy": {"type": "AccessPolicy", "id": self.base_policy_id},
        }

    def post(self):
        """POST method for InheritanceSettings not supported."""
        logging.info("API POST method for InheritanceSettings not supported.")
        pass

    def delete(self, **kwargs):
        """DELETE method for InheritanceSettings not supported."""
        logging.info("API DELETE method for InheritanceSettings not supported.")
        pass



import logging
import fmcapi
import time


def test__subinterfaces(fmc):
    logging.info(
        "Test SubInterfaces.  get, post, put, delete SubInterfaces Objects. Requires registered device"
    )

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    sz1 = fmcapi.SecurityZones(fmc=fmc)
    sz1.name = "_sz1" + namer
    sz1.post()
    time.sleep(1)
    sz2 = fmcapi.SecurityZones(fmc=fmc)
    sz2.name = "_sz2" + namer
    sz2.post()
    time.sleep(1)

    sub1 = fmcapi.SubInterfaces(fmc=fmc, device_name="device_name")
    sub1.p_interface(p_interface="GigabitEthernet0/3", device_name="device_name")
    sub1.enabled = True
    sub1.ifname = "_sub1" + namer
    sub1.subIntfId = "300"
    sub1.vlanId = "300"
    sub1.static(ipv4addr="192.0.2.1", ipv4mask=24)
    sub1.sz(name=sz1.name)
    sub1.post()
    logging.info(sub1.format_data())
    time.sleep(2)

    sub1.get()
    sub1.enabled = False
    sub1.sz(name=sz2.name)
    sub1.put()
    time.sleep(1)

    sub1.get()
    sub1.delete()
    sz1.delete()
    sz2.delete()

    logging.info("Testing SubInterfaces class done.\n")



import logging
import fmcapi
import time
import json


def test__objects_get_query_filters(fmc):
    logging.info("Test Object Get Query Filters.  Get Objects that support their various get query filters.")

    #'_' in nameOrValue='_' is used as an example as we use underscores in our object names. 

    logging.info('Object: Networks')
    obj1 = fmcapi.Networks(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: Hosts')
    obj1 = fmcapi.Hosts(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: NetworkGroups')
    obj1 = fmcapi.NetworkGroups(fmc=fmc)
    groups = obj1.get(nameOrValue='_')
    print(json.dumps(groups, indent=2))
    del obj1

    logging.info('Object: Ports')
    obj1 = fmcapi.Ports(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: ICMPv4Objects')
    obj1 = fmcapi.ICMPv4Objects(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: ICMPv6Objects')
    obj1 = fmcapi.ICMPv6Objects(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: ProtocolPortObjects')
    obj1 = fmcapi.ProtocolPortObjects(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: PortObjectGroups')
    obj1 = fmcapi.PortObjectGroups(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: VlanTags')
    obj1 = fmcapi.VlanTags(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: VlanGroupTags')
    obj1 = fmcapi.VlanGroupTags(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: URLs')
    obj1 = fmcapi.URLs(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: URLGroups')
    obj1 = fmcapi.URLGroups(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: NetworkAddresses')
    obj1 = fmcapi.NetworkAddresses(fmc=fmc)
    obj1.expanded = True
    addresses = obj1.get(nameOrValue='_')
    print(json.dumps(addresses, indent=2))
    del obj1

    logging.info('Object: Ranges')
    obj1 = fmcapi.Ranges(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: FQDNs')
    obj1 = fmcapi.FQDNS(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info('Object: AnyProtocolPortObjects')
    obj1 = fmcapi.AnyProtocolPortObjects(fmc=fmc)
    obj1.get(nameOrValue='_')
    del obj1

    logging.info("Test Object Get Query Filters done.\n")


"""NAT Rules Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftdnatpolicies import FTDNatPolicies
import logging


class NatRules(APIClassTemplate):
    """The NatRules Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    PREFIX_URL = "/policy/ftdnatpolicies"
    REQUIRED_FOR_GET = []
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    def __init__(self, fmc, **kwargs):
        """
        Initialize NatRules object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for NatRules class.")
        self.parse_kwargs(**kwargs)

    def nat_policy(self, name=None, id=None):
        """
        Associate NAT Policy.

        :param name: (str) Name of NAT Policy.
        :param id: (str) ID of NAT Policy.
        :return: None
        """
        logging.debug("In nat_policy() for NatRules class.")
        if id:
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{id}/natrules"
            self.nat_added_to_url = True
        elif name:
            ftd_nat = FTDNatPolicies(fmc=self.fmc)
            ftd_nat.get(name=name)
            if "id" in ftd_nat.__dict__:
                self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{ftd_nat.id}/natrules"
                self.nat_added_to_url = True
            else:
                logging.warning(
                    f"FTD NAT Policy {name} not found.  Cannot set up NatRules for NAT Policy."
                )
        else:
            logging.error("No NatRules name or ID was provided.")

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for NatRules class.")
        if "nat_id" in kwargs:
            self.nat_policy(id=kwargs["nat_id"])
        elif "nat_name" in kwargs:
            self.nat_policy(name=kwargs["nat_name"])
        elif "name" in kwargs:
            self.nat_policy(name=kwargs["name"])

    def post(self):
        """POST method for API for NatRules not supported."""
        logging.info("POST method for API for NatRules not supported.")
        pass

    def put(self):
        """PUT method for API for NatRules not supported."""
        logging.info("PUT method for API for NatRules not supported.")
        pass

    def delete(self):
        """DELETE method for API for NatRules not supported."""
        logging.info("DELETE method for API for NatRules not supported.")
        pass



"""VLAN Tags Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.helper_functions import validate_vlans
import logging


class VlanTags(APIClassTemplate):
    """The VlanTags Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type", "data", "description"]
    VALID_GET_FILTERS = [
        "unusedOnly",
        "nameOrValue",
    ]  # unusedOnly:Bool, nameOrValue:String
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/vlantags"
    REQUIRED_FOR_POST = ["name", "data"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize VlanTags object.

        Set self.type to VlanTag and parse kwargs.
        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for VlanTags class.")
        self.type = "VlanTag"
        self.parse_kwargs(**kwargs)

    def vlans(self, start_vlan, end_vlan=""):
        """
        Associate VLANs.

        :param start_vlan: (int) Lower VLAN.
        :param end_vlan: (int) Upper VLAN.
        """
        logging.debug("In vlans() for VlanTags class.")
        start_vlan, end_vlan = validate_vlans(start_vlan=start_vlan, end_vlan=end_vlan)
        self.data = {"startTag": start_vlan, "endTag": end_vlan}



"""Super class(es) that is inherited by all API objects."""
from .helper_functions import syntax_correcter
import logging
import json

logging.debug(f"In the {__name__} module.")


class APIClassTemplate(object):
    """The base framework for all/(most of) the objects in the FMC."""

    REQUIRED_FOR_POST = ["name"]
    REQUIRED_FOR_PUT = ["id"]
    REQUIRED_FOR_DELETE = ["id"]
    REQUIRED_FOR_GET = [""]
    REQUIRED_GET_FILTERS = []
    FILTER_BY_NAME = False
    URL = ""
    URL_SUFFIX = ""
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\-]"""
    FIRST_SUPPORTED_FMC_VERSION = "6.1"
    VALID_JSON_DATA = []
    VALID_GET_FILTERS = []
    GLOBAL_VALID_FOR_KWARGS = ["dry_run"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []

    @property
    def show_json(self):
        """
        json.dumps of format_data() info.

        :return (str)
        """
        return json.dumps(self.format_data())

    def __init__(self, fmc, **kwargs):
        """
        Initialize an instances of object being created.

        :param fmc: (object) FMC object
        :param **kwargs: Passed variables that will be added to object being instantiated.
        :return: None
        """
        logging.debug("In __init__() for APIClassTemplate class.")
        self.VALID_FOR_KWARGS = self.VALID_FOR_KWARGS + self.GLOBAL_VALID_FOR_KWARGS
        self.fmc = fmc
        self.limit = self.fmc.limit
        self.description = "Created by fmcapi."
        self.overridable = False
        self.dry_run = False
        self.expanded = False
        self.get_filters = {}
        self.URL = f"{self.fmc.configuration_url}{self.URL_SUFFIX}"
        if self.fmc.serverVersion < self.FIRST_SUPPORTED_FMC_VERSION:
            logging.warning(
                f"This API feature was released in version {self.FIRST_SUPPORTED_FMC_VERSION}.  "
                f"Your FMC version is {self.fmc.serverVersion}.  Upgrade to use this feature."
            )

    def format_data(self, filter_query=""):
        """
        Gather all the data in preparation for sending to API in JSON format.

        :param filter_query: (str) 'all' or 'kwargs'
        :return: (dict) json_data
        """
        logging.debug("In format_data() for APIClassTemplate class.")
        json_data = {}
        filter_list = self.VALID_JSON_DATA
        if filter_query == "all":
            filter_list = self.__dict__
        elif filter_query == "kwargs":
            filter_list = self.VALID_FOR_KWARGS
        for key_value in filter_list:
            if hasattr(self, key_value):
                json_data[key_value] = getattr(self, key_value)
        return json_data

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        logging.debug("In parse_kwargs() for APIClassTemplate class.")
        for key_value in self.VALID_FOR_KWARGS:
            if key_value in kwargs:
                if key_value in self.VALID_GET_FILTERS:
                    self.get_filters[key_value] = kwargs[key_value]
                else:
                    self.__dict__[key_value] = kwargs[key_value]
        if "name" in kwargs:
            self.name = syntax_correcter(
                kwargs["name"], permitted_syntax=self.VALID_CHARACTERS_FOR_NAME
            )
            if self.name != kwargs["name"]:
                logging.info(
                    f"Adjusting name '{kwargs['name']}' to '{self.name}' due to invalid characters."
                )

    def valid_for_get(self):
        """
        Use REQUIRED_FOR_GET to ensure all necessary variables exist prior to submitting to API.

        :return: (boolean)
        """
        logging.debug("In valid_for_get() for APIClassTemplate class.")
        if len(self.REQUIRED_GET_FILTERS) > 0:
            for item in self.REQUIRED_GET_FILTERS:
                if item not in self.get_filters:
                    logging.error(
                        f'Missing REQUIRED_GET_FILTERS "{item}" for GET request.'
                    )
                    return False
        if self.REQUIRED_FOR_GET == [""]:
            return True
        for item in self.REQUIRED_FOR_GET:
            if item not in self.__dict__:
                logging.error(f'Missing value "{item}" for GET request.')
                return False
        return True

    def get(self, **kwargs):
        """
        Prepare to send GET call to FMC API.

        If no self.name or self.id exists then return a full listing of all
        objects of this type otherwise return requested name/id values.  Set "expanded=true" results for specific object
        to gather additional detail. Set "unusedOnly=True" to query for unused objects only for certain object types. Set
        "nameOrValue=String" to filter for a particular name or value of an object. This includes partial matches and is
        available for some objects.

        :param: expanded=Bool
        :param: unusedOnly=Bool
        :param: nameOrValue=String
        :return: requests response
        """
        logging.debug("In get() for APIClassTemplate class.")
        self.parse_kwargs(**kwargs)
        if self.fmc.serverVersion < self.FIRST_SUPPORTED_FMC_VERSION:
            logging.error(
                f"Your FMC version, {self.fmc.serverVersion} does not support GET of this feature."
            )
            return {"items": []}
        if self.valid_for_get():
            if "id" in self.__dict__ or "targetId" in self.__dict__:
                if "targetId" in self.__dict__:
                    url = f"{self.URL}/{self.targetId}"
                    if "backupVersion" in self.__dict__:
                        url += f"?backupVersion={self.backupVersion}"
                else:
                    url = f"{self.URL}/{self.id}"
                if self.dry_run:
                    logging.info(
                        "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                    )
                    logging.info("\tMethod = GET")
                    logging.info(f"\tURL = {self.URL}")
                    return False
                response = self.fmc.send_to_api(method="get", url=url)
                try:
                    self.parse_kwargs(**response)
                except TypeError as e:
                    logging.error(
                        f"Response from FMC GET with 'id', {self.id}, returned none."
                        f"That 'id' probably was deleted.  Error: {e}."
                    )
                if "name" in self.__dict__:
                    logging.info(
                        f'GET success. Object with name: "{self.name}" and id: "{self.id}" fetched from FMC.'
                    )
                elif "targetId" in self.__dict__:
                    if "backupVersion" in self.__dict__:
                        logging.info(
                            f'GET success. Object with targetId: "{self.targetId}" backupVersion: "{self.backupVersion}" fetched from FMC.'
                        )
                    logging.info(
                        f'GET success. Object with targetId: "{self.targetId}" fetched from FMC.'
                    )
                else:
                    logging.info(
                        f'GET success. Object with id: "{self.id}" fetched from FMC.'
                    )
            elif "name" in self.__dict__:
                if self.FILTER_BY_NAME:
                    url = f"{self.URL}?name={self.name}&expanded=true"
                else:
                    url = f"{self.URL}?expanded=true"
                    if "limit" in self.__dict__:
                        url = f"{url}&limit={self.limit}"
                    if "offset" in self.__dict__:
                        url = f"{url}&offset={self.offset}"
                response = self.fmc.send_to_api(method="get", url=url)
                if "items" not in response:
                    response["items"] = []
                for item in response["items"]:
                    if "name" in item:
                        if item["name"] == self.name:
                            self.id = item["id"]
                            self.parse_kwargs(**item)
                            logging.info(
                                f'GET success. Object with name: "{self.name}" and id: "{self.id}" '
                                f"fetched from FMC."
                            )
                            return item
                    else:
                        logging.warning(
                            f'No "name" attribute associated with this item to check against {self.name}.'
                        )
                if "id" not in self.__dict__:
                    logging.warning(f"\tGET query for {self.name} is not found.")
                    logging.debug(
                        f"\tGET query for {self.name} is not found.\n\t\tResponse: {json.dumps(response)}"
                    )
            elif len(self.get_filters) > 0:
                url_filter = ""
                for key, value in self.get_filters.items():
                    # Filter value must not be empty otherwise will result in a 400 response
                    if value != "":
                        url_filter += f"{key}%3A{value};"
                    else:
                        logging.warning(
                            f"Terminating GET - {self.URL}?expanded={self.expanded}&filter={url_filter}"
                        )
                        logging.warning(f"{key} MUST have a non empty value")
                        return False
                url = f"{self.URL}?expanded={self.expanded}&filter={url_filter}"
                if self.dry_run:
                    logging.info(
                        "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                    )
                    logging.info("\tMethod = GET")
                    logging.info(f"\tURL = {url}")
                    return False
                response = self.fmc.send_to_api(method="get", url=url)
                if "items" not in response:
                    logging.info(
                        f"GET success. No Objects were found with query filter: {self.get_filters}"
                    )
                    return response
                else:
                    response_count = response.get("paging").get("count")
                    logging.info(
                        f"GET success. {response_count} items found that match query filter: {self.get_filters}"
                    )
                    return response
            else:
                logging.debug(
                    "GET query for object with no name or id set.  "
                    "Returning full list of these object types instead."
                )
                url_suffix_start = "?"
                if url_suffix_start in self.URL:
                    url_suffix_start = "&"
                url = f"{self.URL}{url_suffix_start}expanded=true&limit={self.limit}"
                if self.dry_run:
                    logging.info(
                        "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                    )
                    logging.info("\tMethod = GET")
                    logging.info(f"\tURL = {self.URL}")
                    return False
                response = self.fmc.send_to_api(method="get", url=url)
            if "items" not in response:
                response["items"] = []
            return response
        else:
            logging.warning(
                "get() method failed due to failure to pass valid_for_get() test."
            )
            return False

    def valid_for_post(self):
        """
        Use REQUIRED_FOR_POST to ensure all necessary variables exist prior to submitting to API.

        :return: (boolean)
        """
        logging.debug("In valid_for_post() for APIClassTemplate class.")
        for item in self.REQUIRED_FOR_POST:
            if item not in self.__dict__:
                logging.error(f'Missing value "{item}" for POST request.')
                return False
        return True

    def post(self, **kwargs):
        """
        Prepare to send POST call to FMC API.

        :return: requests response
        """
        logging.debug("In post() for APIClassTemplate class.")
        self.parse_kwargs(**kwargs)
        if self.fmc.serverVersion < self.FIRST_SUPPORTED_FMC_VERSION:
            logging.error(
                f"Your FMC version, {self.fmc.serverVersion} does not support POST of this feature."
            )
            return False
        if "id" in self.__dict__:
            logging.info(
                "ID value exists for this object.  Redirecting to put() method."
            )
            self.put()
        else:
            if self.valid_for_post():
                if self.dry_run:
                    logging.info(
                        "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                    )
                    logging.info("\tMethod = POST")
                    logging.info(f"\tURL = {self.URL}")
                    logging.info(f"\tJSON = {self.show_json}")
                    return False
                response = self.fmc.send_to_api(
                    method="post", url=self.URL, json_data=self.format_data()
                )
                if response:
                    self.parse_kwargs(**response)
                    if "name" in self.__dict__ and "id" in self.__dict__:
                        logging.info(
                            f'POST success. Object with name: "{self.name}" and id: "{self.id}" created in FMC.'
                        )
                    else:
                        logging.debug(
                            'POST success but no "id" or "name" values in API response.'
                        )
                else:
                    logging.warning("POST failure.  No data in API response.")
                return response
            else:
                logging.warning(
                    "post() method failed due to failure to pass valid_for_post() test."
                )
                return False

    def valid_for_put(self):
        """
        Use REQUIRED_FOR_PUT to ensure all necessary variables exist prior to submitting to API.

        :return: (boolean)
        """
        logging.debug("In valid_for_put() for APIClassTemplate class.")
        for item in self.REQUIRED_FOR_PUT:
            if item not in self.__dict__:
                logging.error(f'Missing value "{item}" for PUT request.')
                return False
        return True

    def put(self, **kwargs):
        """
        Prepare to send PUT call to FMC API.

        :return: requests response
        """
        logging.debug("In put() for APIClassTemplate class.")
        self.parse_kwargs(**kwargs)
        if self.fmc.serverVersion < self.FIRST_SUPPORTED_FMC_VERSION:
            logging.error(
                f"Your FMC version, {self.fmc.serverVersion} does not support PUT of this feature."
            )
            return False
        if self.valid_for_put():
            url = f"{self.URL}/{self.id}"
            if self.dry_run:
                logging.info(
                    "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                )
                logging.info("\tMethod = PUT")
                logging.info(f"\tURL = {self.URL}")
                logging.info(f"\tJSON = {self.show_json}")
                return False
            response = self.fmc.send_to_api(
                method="put", url=url, json_data=self.format_data()
            )
            self.parse_kwargs(**response)
            if "name" in self.__dict__:
                logging.info(
                    f'PUT success. Object with name: "{self.name}" and id: "{self.id}" updated in FMC.'
                )
            else:
                logging.info(
                    f'PUT success. Object with id: "{self.id}" updated in FMC.'
                )
            return response
        else:
            logging.warning(
                "put() method failed due to failure to pass valid_for_put() test."
            )
            return False

    def valid_for_delete(self):
        """
        Use REQUIRED_FOR_DELETE to ensure all necessary variables exist prior to submitting to API.

        :return: (boolean)
        """
        logging.debug("In valid_for_delete() for APIClassTemplate class.")
        for item in self.REQUIRED_FOR_DELETE:
            if item not in self.__dict__:
                logging.error(f'Missing value "{item}" for DELETE request.')
                return False
        return True

    def delete(self, **kwargs):
        """
        Prepare to send DELETE call to FMC API.

        :return: requests response
        """
        logging.debug("In delete() for APIClassTemplate class.")
        self.parse_kwargs(**kwargs)
        if self.fmc.serverVersion < self.FIRST_SUPPORTED_FMC_VERSION:
            logging.error(
                f"Your FMC version, {self.fmc.serverVersion} does not support DELETE of this feature."
            )
            return False
        if self.valid_for_delete():
            if "targetId" in self.__dict__:
                url = f"{self.URL}/{self.targetId}"
                if "backupVersion" in self.__dict__:
                    url += f"?backupVersion={self.backupVersion}"
            else:
                url = f"{self.URL}/{self.id}"
            if self.dry_run:
                logging.info(
                    "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                )
                logging.info("\tMethod = DELETE")
                logging.info(f"\tURL = {self.URL}")
                logging.info(f"\tJSON = {self.show_json}")
                return False
            response = self.fmc.send_to_api(
                method="delete", url=url, json_data=self.format_data()
            )
            if not response:
                return None
            self.parse_kwargs(**response)
            if hasattr(self, "name"):
                logging.info(
                    f'DELETE success. Object with name: "{self.name}" and id: "{self.id}" deleted in FMC.'
                )
            elif "targetId" in self.__dict__:
                if "backupVersion" in self.__dict__:
                    logging.info(
                        f'DELETE success. Object with targetId: "{self.targetId}" backupVersion: "{self.backupVersion}" deleted from FMC.'
                    )
                else:
                    logging.info(
                        f'DELETE success. Object with targetId: "{self.targetId}" deleted from FMC.'
                    )
            else:
                logging.info(f'DELETE success. Object id: "{self.id}" deleted in FMC.')
            return response
        else:
            logging.warning(
                "delete() method failed due to failure to pass valid_for_delete() test."
            )
            return False



"""FTD NAT Policies Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class FTDNatPolicies(APIClassTemplate):
    """The FTDNatPolicies Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/policy/ftdnatpolicies"
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    def __init__(self, fmc, **kwargs):
        """
        Initialize FTDNatPolicies object.

        Set self.type to "FTDNatPolicy", parse the kwargs, and set up the self.URL.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for FTDNatPolicies class.")
        self.parse_kwargs(**kwargs)
        self.type = "FTDNatPolicy"



"""Security Zones Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class SecurityZones(APIClassTemplate):
    """The SecurityZones Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "description",
        "interfaceMode",
        "interfaces",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/securityzones"
    REQUIRED_FOR_POST = ["name", "interfaceMode"]
    FILTER_BY_NAME = True

    def __init__(self, fmc, **kwargs):
        """
        Initialize SecurityZones object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for SecurityZones class.")
        self.parse_kwargs(**kwargs)

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for SecurityZones class.")
        if "interfaceMode" in kwargs:
            self.interfaceMode = kwargs["interfaceMode"]
        else:
            self.interfaceMode = "ROUTED"



import logging
import fmcapi
import time


def test__hitcounts(fmc, device_name="", prefilter_id=""):
    if not device_name and not prefilter_id:
        return f"Name of an actual device or prefilter ID is required for the HitCounts test to work... skipping test."

    logging.info(
        "In preparation for testing HitCounts method, set up some known objects in the FMC."
    )
    starttime = str(int(time.time()))
    namer = f"test__hitcounts_{starttime}"

    # Get the device
    device1 = fmcapi.DeviceRecords(fmc=fmc, name=device_name)
    device1.get()

    # In case there is no ACP Rule build a temp one.
    acprule1 = fmcapi.AccessRules(fmc=fmc, acp_id=device1.accessPolicy["id"])
    # acprule1 = fmcapi.AccessRules(fmc=fmc, acp_name=device1.accessPolicy['name'])
    acprule1.name = namer
    acprule1.action = "ALLOW"
    acprule1.post()
    time.sleep(1)
    acprule1.get()

    hitcounter1 = None
    if prefilter_id:
        hitcounter1 = fmcapi.HitCounts(
            fmc=fmc, prefilter_id=prefilter_id, device_name=device_name
        )
    else:
        hitcounter1 = fmcapi.HitCounts(
            fmc=fmc, acp_id=device1.accessPolicy["id"], device_name=device_name
        )
        """ 
        Searching for AccessRule by name returns the "correct" ID for the rule but HitCount shows a completely
        different ID so it doesn't match.
        # hitcounter1.acp_rules(action="add", name="Permit HQ LAN")

        If you know the ID that HitCount is looking for a specific rule:
        # hitcounter1.acp_rules(action="add", acp_rule_id="005056B5-44E6-0ed3-0000-000268434433")
        """
    for result in hitcounter1.get():
        print(result)

    logging.info("Test HitCount done.")

    logging.info("Cleanup of testing HitCount methods.")
    acprule1.delete()
    time.sleep(1)
    logging.info("Cleanup of objects for HitCount test done.\n\n")



"""RAVpn Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class RAVpn(APIClassTemplate):
    """The RAVpn Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "connectionProfiles",
        "groupPolicies",
        "accessInterfaceSettings",
        "ikev2Policies",
        "addressAssignmentSettings",
        "anyConnectClientImages",
        "externalBrowserPackage",
        "ipsecCryptoMaps",
        "certificateMapSettings",
        "ipsecAdvancedSettings",
        "secureClientCustomizationSettings",
        "dapPolicy",
        "ldapAttributeMaps",
        "loadBalanceSettings",
        "configureSSL",
        "configureIpsec",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/policy/ravpns"
    FIRST_SUPPORTED_FMC_VERSION = "7.2"

    def __init__(self, fmc, **kwargs):
        """
        Initialize RAVpn object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for RAVPN class.")
        self.type = "RAVpn"
        self.parse_kwargs(**kwargs)

    # POST, PUT and DELETE are valid endpoints in the api >7.2, however ravpn POST has a cycical dependency
    # with serveral other api endpoints - connectionprofiles, addressassignments, etc.
    # This is due to needing a container uuid of the RAVPN policy but also needing those objects to
    # create the RAVPN policy in the first place.

    def post(self):
        """POST method for API for RAVPN not supported."""
        logging.info("POST method for API for RAVPN not supported.")
        pass



import logging
import fmcapi


def test__application(fmc):
    logging.info("Testing Application class.")

    obj1 = fmcapi.Applications(fmc=fmc)
    logging.info("All Application -- >")
    result = obj1.get(limit=1000)
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")

    del obj1
    obj1 = fmcapi.Applications(fmc=fmc, name="WD softwares Download/Update")
    logging.info("One Application -- >")
    logging.info(obj1.get(limit=1000))

    logging.info("Testing Application class done.\n")



"""VLAN Group Tags Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .vlantags import VlanTags
from fmcapi.api_objects.helper_functions import validate_vlans
import logging


class VlanGroupTags(APIClassTemplate):
    """The VlanGroupTags Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type", "description", "objects", "literals"]
    VALID_GET_FILTERS = [
        "unusedOnly",
        "nameOrValue",
    ]  # unusedOnly:Bool, nameOrValue:String
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/vlangrouptags"

    # Technically you can have objects OR literals but I'm not set up for "OR" logic, yet.
    REQUIRED_FOR_POST = ["name", "objects"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize VlanGroupTags object.

        Set self.type to VlanGroupTag and parse kwargs.
        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for VlanGroupTags class.")
        self.parse_kwargs(**kwargs)
        self.type = "VlanGroupTag"

    def named_vlantags(self, action, name=""):
        """
        Associate Named VLAN Tags.

        :param action: (str) 'add', 'remove', or 'clear'.
        :param name: (str) Name of VLAN Tags.
        """
        logging.debug("In named_vlantags() for VlanGroupTags class.")
        if action == "add":
            vlan1 = VlanTags(fmc=self.fmc)
            response = vlan1.get()
            if "items" in response:
                new_vlan = None
                for item in response["items"]:
                    if item["name"] == name:
                        new_vlan = {
                            "name": item["name"],
                            "id": item["id"],
                            "type": item["type"],
                        }
                        break
                if new_vlan is None:
                    logging.warning(
                        f'VlanTag "{name}" is not found in FMC.  Cannot add to VlanGroupTags.'
                    )
                else:
                    if "objects" in self.__dict__:
                        duplicate = False
                        for obj in self.objects:
                            if obj["name"] == new_vlan["name"]:
                                duplicate = True
                                break
                        if not duplicate:
                            self.objects.append(new_vlan)
                            logging.info(f'Adding "{name}" to VlanGroupTags.')
                    else:
                        self.objects = [new_vlan]
                        logging.info(f'Adding "{name}" to VlanGroupTags.')
        elif action == "remove":
            if "objects" in self.__dict__:
                objects_list = []
                for obj in self.objects:
                    if obj["name"] != name:
                        objects_list.append(obj)
                self.objects = objects_list
                logging.info(f'Removed "{name}" from VlanGroupTags.')
            else:
                logging.info(
                    "This VlanGroupTags has no named_vlantags.  Nothing to remove."
                )
        elif action == "clear":
            if "objects" in self.__dict__:
                del self.objects
                logging.info("All named_vlantags removed from this VlanGroupTags.")

    def unnamed_vlantags(self, action, startvlan="", endvlan=""):
        """
        Associate Unnamed VLAN Tags.

        :param action: (str) 'add', 'remove', or 'clear'
        :param startvlan: (int) Lower VLAN.
        :param endvlan: (int) Upper VLAN.
        """
        logging.debug("In unnamed_vlantags() for VlanGroupTags class.")
        if action == "add":
            startvlan, endvlan = validate_vlans(start_vlan=startvlan, end_vlan=endvlan)
            new_literal = {"startTag": startvlan, "endTag": endvlan, "type": ""}
            if "literals" in self.__dict__:
                duplicate = False
                for obj in self.literals:
                    if (
                        obj["startTag"] == new_literal["startTag"]
                        and obj["endTag"] == new_literal["endTag"]
                    ):
                        duplicate = True
                        break
                if not duplicate:
                    self.literals.append(new_literal)
                    logging.info(f'Adding "{startvlan}/{endvlan}" to VlanGroupTags.')
            else:
                self.literals = [new_literal]
                logging.info(f'Adding "{startvlan}/{endvlan}" to VlanGroupTags.')
        elif action == "remove":
            startvlan, endvlan = validate_vlans(start_vlan=startvlan, end_vlan=endvlan)
            if "literals" in self.__dict__:
                literals_list = []
                for obj in self.literals:
                    if obj["startTag"] != startvlan and obj["endTag"] != endvlan:
                        literals_list.append(obj)
                self.literals = literals_list
                logging.info(f'Removed "{startvlan}/{endvlan}" from VlanGroupTags.')
            else:
                logging.info(
                    "This VlanGroupTag has no unnamed_vlantags.  Nothing to remove."
                )
        elif action == "clear":
            if "literals" in self.__dict__:
                del self.literals
                logging.info("All unnamed_vlantags removed from this VlanGroupTags.")



"""URL Groups Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .urls import URLs
import logging


class URLGroups(APIClassTemplate):
    """The URLGroups Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type", "objects", "literals"]
    VALID_GET_FILTERS = [
        "unusedOnly",
        "nameOrValue",
    ]  # unusedOnly:Bool, nameOrValue:String
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/urlgroups"

    # Technically you can have objects OR literals but I'm not set up for "OR" logic, yet.
    REQUIRED_FOR_POST = ["name", "objects"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize URLGroups object.

        Set self.type to "URLGroup" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for URLGroups class.")
        self.parse_kwargs(**kwargs)
        self.type = "URLGroup"

    def named_urls(self, action, name=""):
        """
        Associate Named URLs.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of URL.
        """
        logging.debug("In named_urls() for URLGroups class.")
        if action == "add":
            url1 = URLs(fmc=self.fmc)
            response = url1.get()
            if "items" in response:
                new_url = None
                for item in response["items"]:
                    if item["name"] == name:
                        new_url = {
                            "name": item["name"],
                            "id": item["id"],
                            "type": item["type"],
                        }
                        break
                if new_url is None:
                    logging.warning(
                        f'URL "{name}" is not found in FMC.  Cannot add to URLGroups.'
                    )
                else:
                    if "objects" in self.__dict__:
                        duplicate = False
                        for obj in self.objects:
                            if obj["name"] == new_url["name"]:
                                duplicate = True
                                break
                        if not duplicate:
                            self.objects.append(new_url)
                            logging.info(f'Adding "{name}" to URLGroups.')
                    else:
                        self.objects = [new_url]
                        logging.info(f'Adding "{name}" to URLGroups.')
        elif action == "remove":
            if "objects" in self.__dict__:
                objects_list = []
                for obj in self.objects:
                    if obj["name"] != name:
                        objects_list.append(obj)
                self.objects = objects_list
                logging.info(f'Removed "{name}" from URLGroups.')
            else:
                logging.info("This URLGroups has no named_urls.  Nothing to remove.")
        elif action == "clear":
            if "objects" in self.__dict__:
                del self.objects
                logging.info("All named_urls removed from this URLGroups.")

    def unnamed_urls(self, action, value=""):
        """
        Associate Unnamed URLs.

        :param action: (str) 'add', 'remove', or 'clear'
        :param value: (str) URL.
        """
        logging.debug("In unnamed_urls() for URLGroups class.")
        if action == "add":
            if value == "":
                logging.error(
                    "Value assignment required to add unnamed_url to URLGroups."
                )
                return
            value_type = "Url"
            new_literal = {"type": value_type, "url": value}
            if "literals" in self.__dict__:
                duplicate = False
                for obj in self.literals:
                    if obj["url"] == new_literal["url"]:
                        duplicate = True
                        break
                if not duplicate:
                    self.literals.append(new_literal)
                    logging.info(f'Adding "{value}" to URLGroups.')
            else:
                self.literals = [new_literal]
                logging.info(f'Adding "{value}" to URLGroups.')
        elif action == "remove":
            if "literals" in self.__dict__:
                literals_list = []
                for obj in self.literals:
                    if obj["url"] != value:
                        literals_list.append(obj)
                self.literals = literals_list
                logging.info(f'Removed "{value}" from URLGroups.')
            else:
                logging.info("This URLGroups has no unnamed_urls.  Nothing to remove.")
        elif action == "clear":
            if "literals" in self.__dict__:
                del self.literals
                logging.info("All unnamed_urls removed from this URLGroups.")



"""Deployment Services Classes."""

import logging
from .deployabledevices import DeployableDevices
from .deploymentrequests import DeploymentRequests

logging.debug("In the deployment_services __init__.py file.")

__all__ = ["DeployableDevices", "DeploymentRequests"]



"""Interface Groups Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.device_services.physicalinterfaces import PhysicalInterfaces
import logging


class InterfaceGroups(APIClassTemplate):
    """The InterfaceGroups Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "description", "interfaceMode", "interfaces"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/interfacegroups"
    REQUIRED_FOR_POST = ["name", "interfaceMode"]
    REQUIRED_FOR_PUT = ["id"]
    FILTER_BY_NAME = True

    def __init__(self, fmc, **kwargs):
        """
        Initialize InterfaceGroups object.

        Set self.type to "InterfaceGroup" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for InterfaceGroups class.")
        self.parse_kwargs(**kwargs)
        self.type = "InterfaceGroup"

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for InterfaceGroups class.")
        if "interfaceMode" in kwargs:
            self.interfaceMode = kwargs["interfaceMode"]
        else:
            self.interfaceMode = "ROUTED"

    def p_interface(self, device_name="", action="add", names=[]):
        """
        Associate Physical Interface.

        :param device_name: (str) Name of device.
        :param action: (str) 'add', 'remove', or 'clear'
        :param names: (list) List of interface names
        """
        logging.debug("In interfaces() for InterfaceGroups class.")
        if action == "add":
            intfs = []
            for name in names:
                intf = PhysicalInterfaces(fmc=self.fmc)
                intf.get(name=name, device_name=device_name)
                if "id" in intf.__dict__ and "ifname" in intf.__dict__:
                    intfs.append({"name": intf.name, "id": intf.id, "type": intf.type})
                elif "id" in intf.__dict__:
                    logging.warning(
                        f'PhysicalInterface, "{name}", found without logical ifname.  '
                        f"Cannot add to InterfaceGroups."
                    )
                else:
                    logging.warning(
                        f'PhysicalInterface, "{name}", not found.  Cannot add to InterfaceGroups.'
                    )
            if len(intfs) != 0:
                # Make sure we found at least one intf
                self.interfaces = intfs
            else:
                logging.warning(
                    f'No valid PhysicalInterface found: "{names}".  Cannot remove from InterfaceGroups.'
                )
        elif action == "remove":
            if "interfaces" in self.__dict__:
                intfs = []
                for interface in self.interfaces:
                    if interface["name"] not in names:
                        intfs.append(interface)
                    else:
                        logging.info(
                            f"""Removed "{interface['name']}" from InterfaceGroups."""
                        )
                self.interfaces = intfs
            else:
                logging.warning(
                    "This InterfaceObject has no interfaces.  Nothing to remove."
                )
        elif action == "clear" or action == "clear-all":
            if "interfaces" in self.__dict__:
                del self.interfaces
                logging.info(
                    "All PhysicalInterfaces removed from this InterfaceGroups."
                )



import logging
import fmcapi
import time


def test__bridge_group_interfaces(fmc):
    logging.info(
        "Test BridgeGroupInterfaces.  get, post, put, delete BridgeGroupInterfaces Objects. "
        "Requires registered device"
    )

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    sz1 = fmcapi.SecurityZones(fmc=fmc)
    sz1.name = "_sz1" + namer
    sz1.post()
    time.sleep(1)
    sz2 = fmcapi.SecurityZones(fmc=fmc)
    sz2.name = "_sz2" + namer
    sz2.post()
    time.sleep(1)

    br1 = fmcapi.BridgeGroupInterfaces(fmc=fmc, device_name="device_name")
    br1.p_interfaces(
        p_interfaces=["GigabitEthernet0/3", "GigabitEthernet0/5"],
        device_name="device_name",
    )
    br1.enabled = True
    br1.ifname = "_br1" + namer
    br1.bridgeGroupId = "1"
    br1.static(ipv4addr="192.0.2.1", ipv4mask=24)
    br1.sz(name=sz1.name)
    br1.post()
    time.sleep(2)

    br1.get()
    br1.enabled = False
    br1.sz(name=sz2.name)
    br1.put()
    time.sleep(1)

    br1.get()
    br1.delete()
    sz1.delete()
    sz2.delete()

    logging.info("Testing BridgeGroupInterfaces class done.\n")



"""Update Packages Classes."""

import logging
from .listapplicabledevices import ListApplicableDevices
from .upgradepackages import UpgradePackages
from .upgradepackage import Upgrades

logging.debug("In the update_packages __init__.py file.")

__all__ = [
    "ListApplicableDevices",
    "UpgradePackages",
    "Upgrades",
]



"""FTDs to VPNs Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class FTDS2SVPNs(APIClassTemplate):
    """The FTDS2SVPNs Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "ipsecSettings",
        "endpoints",
        "ikeSettings",
        "advancedSettings",
        "description",
        "ikeV2Enabled",
        "ikeV1Enabled",
        "topologyType",
        "version",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    FIRST_SUPPORTED_FMC_VERSION = "6.3"
    URL_SUFFIX = "/policy/ftds2svpns"
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    def __init__(self, fmc, **kwargs):
        """
        Initialize FTDS2SVPNs object.

        Set self.type to "FTDS2SVPpn", parse the kwargs, and set up the self.URL.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for FTDNatPolicies class.")
        self.parse_kwargs(**kwargs)
        self.type = "FTDS2SVpn"



import logging
import fmcapi
import time


def test__ikev1(fmc):
    logging.info(
        "Test IKEv1Policies and IKEv1IpsecProposals."
        "  Post, get, put, delete IKEv1Policies and IKEv1IpsecProposals Objects."
    )

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    ipsec1 = fmcapi.IKEv1IpsecProposals(fmc=fmc)
    ipsec1.name = "_ipsec" + namer
    ipsec1.espEncryption = "AES-128"
    ipsec1.espHash = "SHA"
    ipsec1.post()

    ipsec1.get()
    ipsec1.espEncryption = "AES-192"
    ipsec1.put()

    pol1 = fmcapi.IKEv1Policies(fmc=fmc)
    pol1.name = "_pol" + namer
    pol1.encryption = "3DES"
    pol1.hash = "SHA"
    pol1.priority = "10"
    pol1.diffieHellmanGroup = "5"
    pol1.authenticationMethod = "Preshared Key"
    pol1.lifetimeInSeconds = "3600"
    pol1.post()

    pol1.get()
    pol1.encryption = "AES-128"
    pol1.put()

    ipsec1.delete()
    pol1.delete()

    logging.info("Test IKEv1Policies and IKEv1IpsecProposals classes done.\n")



import logging
import fmcapi
import time


def test__ikev2(fmc):
    logging.info(
        "Test IKEv2Policies and IKEv2IpsecProposals."
        "  Post, get, put, delete IKEv2Policies and IKEv2IpsecProposals Objects."
    )

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    encryption_list = ["AES", "AES-192", "AES-256", "NULL"]
    integrity_list1 = ["NULL", "SHA-1", "SHA-256", "SHA-384", "SHA-512"]
    ipsec_integrity_list1 = ["NULL", "SHA", "SHA-256", "SHA-384", "SHA-512"]
    # 'NULL' is invalid for prf_integrity.  Should generate a warning log and ignore that type.
    prf_integrity_list1 = ["NULL", "SHA", "SHA-256", "SHA-384", "SHA-512"]

    ipsec1 = fmcapi.IKEv2IpsecProposals(fmc=fmc)
    ipsec1.name = "_ipsec" + namer
    ipsec1.encryption(action="add", algorithms=encryption_list)
    ipsec1.hash(action="add", algorithms=integrity_list1)
    ipsec1.post()

    ipsec1.get()
    # Try to add a duplicate
    ipsec1.encryption(action="add", algorithms=["AES-192"])
    ipsec1.hash(action="add", algorithms=["SHA-1"])

    ipsec1.encryption(action="remove", algorithms=["NULL"])
    ipsec1.hash(action="remove", algorithms=["NULL"])
    ipsec1.put()

    # None of the algorithms can contain an empty list
    ipsec1.get()
    ipsec1.encryption(action="clear")
    ipsec1.hash(action="clear")
    ipsec1.encryption(action="add", algorithms=["AES-192"])
    ipsec1.hash(action="add", algorithms=["SHA-1"])
    ipsec1.put()

    pol1 = fmcapi.IKEv2Policies(fmc=fmc)
    pol1.name = "_pol" + namer
    pol1.priority = "10"
    pol1.diffieHellmanGroups = ["2", "5"]
    pol1.encryption(action="add", algorithms=encryption_list)
    pol1.hash(action="add", algorithms=ipsec_integrity_list1)
    pol1.prf_hash(action="add", algorithms=prf_integrity_list1)
    pol1.lifetimeInSeconds = "3600"
    pol1.post()

    pol1.get()
    # Try to add a duplicate
    pol1.encryption(action="add", algorithms=["AES-192"])
    pol1.hash(action="add", algorithms=["NULL"])
    pol1.prf_hash(action="add", algorithms=["SHA"])

    pol1.encryption(action="remove", algorithms=["NULL"])
    pol1.hash(action="remove", algorithms=["NULL"])
    pol1.prf_hash(action="remove", algorithms=["SHA"])
    pol1.put()

    # None of the algorithms can contain an empty list
    pol1.get()
    pol1.encryption(action="clear")
    pol1.hash(action="clear")
    pol1.prf_hash(action="clear")
    pol1.encryption(action="add", algorithms=["AES-192"])
    pol1.hash(action="add", algorithms=["NULL"])
    pol1.prf_hash(action="add", algorithms=["SHA"])
    pol1.put()

    ipsec1.delete()
    pol1.delete()

    logging.info("Test IKEv2Policies and IKEv2IpsecProposals classes done.\n")



"""FTDDeviceHAPairs Classes."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.device_services.devicerecords import DeviceRecords
import logging


class FTDDeviceHAPairs(APIClassTemplate):
    """The FTDDeviceHAPairs Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "primary",
        "secondary",
        "ftdHABootstrap",
        "action",
        "forceBreak",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/devicehapairs/ftddevicehapairs"
    REQUIRED_FOR_POST = ["primary", "secondary", "ftdHABootstrap"]
    REQUIRED_FOR_PUT = ["id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize FTDDeviceHAPairs object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for FTDDeviceHAPairs class.")
        self.parse_kwargs(**kwargs)

    def device(self, primary_name="", secondary_name=""):
        """
        Add primary and secondary devices to HA Pair.

        :param primary_name: (str) 1st Device.
        :param secondary_name: (str) 2nd Device.
        :return: None
        """
        logging.debug("In device() for FTDDeviceHAPairs class.")
        primary = DeviceRecords(fmc=self.fmc)
        primary.get(name=primary_name)
        secondary = DeviceRecords(fmc=self.fmc)
        secondary.get(name=secondary_name)
        if "id" in primary.__dict__:
            self.primary_id = primary.id
        else:
            logging.warning(
                f"Device {primary_name} not found.  Cannot set up device for FTDDeviceHAPairs."
            )
        if "id" in secondary.__dict__:
            self.secondary_id = secondary.id
        else:
            logging.warning(
                f"Device {secondary_name} not found.  Cannot set up device for FTDDeviceHAPairs."
            )

    def primary(self, name):
        """
        Identify primary device.

        :param name: (str) Name of primary device.
        :return: None
        """
        logging.debug("In primary() for FTDDeviceHAPairs class.")
        primary = DeviceRecords(fmc=self.fmc)
        primary.get(name=name)
        if "id" in primary.__dict__:
            self.primary = {"id": primary.id}
        else:
            logging.warning(
                f"Device {primary.name} not found.  Cannot set up device for FTDDeviceHAPairs."
            )

    def secondary(self, name):
        """
        Identify secondary device.

        :param name: (str) Name of secondary device.
        :return: None
        """
        logging.debug("In secondary() for DeviceHAPairs class.")
        secondary = DeviceRecords(fmc=self.fmc)
        secondary.get(name=name)
        if "id" in secondary.__dict__:
            self.secondary = {"id": secondary.id}
        else:
            logging.warning(
                f"Device {secondary.name} not found.  Cannot set up device for FTDDeviceHAPairs."
            )

    def switch_ha(self):
        """
        Set up HA SWITCH.

        :return: None
        """
        logging.debug("In switch_ha() for FTDDeviceHAPairs class.")
        ha1 = DeviceHAPairs(fmc=self.fmc)
        ha1.get(name=self.name)
        if "id" in ha1.__dict__:
            self.id = ha1.id
            self.action = "SWITCH"
        else:
            logging.warning(
                f"FTDDeviceHAPairs {self.name} not found.  Cannot set up HA for SWITCH."
            )

    def break_ha(self):
        """
        Destroy HA pair.

        :return: None
        """
        logging.debug("In break_ha() for FTDDeviceHAPairs class.")
        ha1 = DeviceHAPairs(fmc=self.fmc)
        ha1.get(name=self.name)
        if "id" in ha1.__dict__:
            self.id = ha1.id
            self.action = "HABREAK"
            self.forceBreak = True
        else:
            logging.warning(
                f"FTDDeviceHAPairs {self.name} not found.  Cannot set up HA for BREAK."
            )

    def post(self, **kwargs):
        """
        Disable autodeploy and then POST.

        :return: requests response
        """
        logging.debug("In post() for FTDDeviceHAPairs class.")
        # Attempting to "Deploy" during Device registration causes issues.
        self.fmc.autodeploy = False
        return super().post(**kwargs)

    def put(self, **kwargs):
        """
        Disable autodeploy and then PUT.

        :return: requests response
        """
        logging.debug("In put() for FTDDeviceHAPairs class.")
        # Attempting to "Deploy" during Device registration causes issues.
        self.fmc.autodeploy = False
        return super().put(**kwargs)



"""Advanced Settings Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftds2svpns import FTDS2SVPNs
import logging


class AdvancedSettings(APIClassTemplate):
    """The AdvancedSettings Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "advancedIkeSetting",
        "advancedTunnelSetting",
        "advancedIpsecSetting",
        "version",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    FIRST_SUPPORTED_FMC_VERSION = "6.3"
    PREFIX_URL = "/policy/ftds2svpns"
    REQUIRED_FOR_POST = ["vpn_id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize AdvancedSettings object.

        :param fmc: (object) FMC object
        :param **kwargs: Set initial variables during instantiation of AdvancedSettings object.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for AdvancedSettings class.")
        self.parse_kwargs(**kwargs)
        self.type = "AdvancedSettings"

    def vpn_policy(self, pol_name):
        """
        Associate a Policy with this VPN.

        :param pol_name: (str) Name of policy.
        :return: None
        """
        logging.debug("In vpn_policy() for AdvancedSettings class.")
        ftd_s2s = FTDS2SVPNs(fmc=self.fmc)
        ftd_s2s.get(name=pol_name)
        if "id" in ftd_s2s.__dict__:
            self.vpn_id = ftd_s2s.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.vpn_id}/advancedsettings"
            self.vpn_added_to_url = True
        else:
            logging.warning(
                f'FTD S2S VPN Policy "{pol_name}" not found.  Cannot set up AdvancedSettings for FTDS2SVPNs Policy.'
            )



"""Device Services Classes."""

import logging
from .bridgegroupinterfaces import BridgeGroupInterfaces
from .devicerecords import DeviceRecords
from .etherchannelinterfaces import EtherchannelInterfaces
from .ipv4staticroutes import IPv4StaticRoutes
from .ipv6staticroutes import IPv6StaticRoutes
from .physicalinterfaces import PhysicalInterfaces
from .redundantinterfaces import RedundantInterfaces
from .staticroutes import StaticRoutes
from .subinterfaces import SubInterfaces

logging.debug("In the device_services __init__.py file.")

__all__ = [
    "DeviceRecords",
    "StaticRoutes",
    "IPv4StaticRoutes",
    "IPv6StaticRoutes",
    "PhysicalInterfaces",
    "BridgeGroupInterfaces",
    "RedundantInterfaces",
    "EtherchannelInterfaces",
    "SubInterfaces",
]



import logging
import fmcapi
import time


def test__vlan_group_tag(fmc):
    logging.info("Testing VlanGroupTag class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj10 = fmcapi.VlanTags(
        fmc=fmc, name="_vlantag10", data={"startTag": "888", "endTag": "999"}
    )
    obj10.post()
    obj11 = fmcapi.VlanTags(
        fmc=fmc, name="_vlantag11", data={"startTag": "222", "endTag": "333"}
    )
    obj11.post()
    obj12 = fmcapi.VlanTags(
        fmc=fmc, name="_vlantag12", data={"startTag": "1", "endTag": "999"}
    )
    obj12.post()
    time.sleep(1)
    obj1 = fmcapi.VlanGroupTags(fmc=fmc, name=namer)
    obj1.named_vlantags(action="add", name=obj10.name)
    obj1.named_vlantags(action="add", name=obj11.name)
    obj1.named_vlantags(action="remove", name=obj11.name)
    obj1.named_vlantags(action="clear")
    obj1.named_vlantags(action="add", name=obj10.name)
    obj1.named_vlantags(action="add", name=obj11.name)
    obj1.named_vlantags(action="add", name=obj12.name)
    obj1.named_vlantags(action="remove", name=obj12.name)
    obj1.post()
    time.sleep(1)
    del obj1
    obj1 = fmcapi.VlanGroupTags(fmc=fmc, name=namer)
    obj1.get()
    obj1.unnamed_vlantags(action="add", startvlan="22", endvlan="33")
    obj1.unnamed_vlantags(action="clear")
    obj1.unnamed_vlantags(action="add", startvlan="22", endvlan="33")
    obj1.unnamed_vlantags(action="remove", startvlan="22", endvlan="33")
    obj1.unnamed_vlantags(action="add", startvlan="44", endvlan="33")
    obj1.unnamed_vlantags(action="add", startvlan="900")
    obj1.put()
    time.sleep(1)
    obj1.delete()
    obj10.delete()
    obj11.delete()
    obj12.delete()

    logging.info("Testing VlanGroupTag class done.\n")



import logging
import fmcapi


def test__deployable_devices(fmc):
    logging.info("Testing DeployableDevices() class.")
    tmp = fmcapi.DeployableDevices(fmc=fmc)
    logging.info(tmp.get())
    logging.info("Testing DeployableDevices() method done.\n")



"""Device Group Services Classes."""

import logging
from .devicegrouprecords import DeviceGroupRecords

logging.debug("In the device_group_services __init__.py file.")

__all__: [
    "DeviceGroupRecords",
]



"""ConnectionProfiles Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class ConnectionProfiles(APIClassTemplate):
    """The ConnectionProfiles Object in the FMC."""

    REQUIRED_FOR_GET = ["container_uuid"]
    REQUIRED_FOR_POST = ["container_uuid", "name"]
    REQUIRED_FOR_PUT = ["container_uuid", "id"]
    REQUIRED_FOR_DELETE = ["container_uuid", "id"]
    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "groupPolicy",
        "authenticationMethod",
        "ipv4AddressPool",
        "enableSecondaryAuthentication",
        "enableSecondaryAuthFallbackToLocal",
        "useLocalAsSecondaryAuthServer",
        "enablePrimaryAuthFallbackToLocal",
        "useLocalAsPrimaryAuthServer",
        "primaryAuthenticationServer",
        "allowConnectionOnlyIfAuthorized",
        "enablePasswordManagement",
        "stripGroupFromUsername",
        "stripRealmFromUsername",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + ["container_uuid"]
    URL_SUFFIX = "/policy/ravpns"
    FIRST_SUPPORTED_FMC_VERSION = "7.2"

    def __init__(self, fmc, **kwargs):
        """
        Initialize ConnectionProfiles object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for ConnectionProfiles class.")
        self.type = "ConnectionProfiles"
        self.parse_kwargs(**kwargs)
        URL_CONTAINER_SUFFIX = f"/{self.container_uuid}/connectionprofiles"
        self.URL = self.URL + URL_CONTAINER_SUFFIX



"""Extended Access List Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class ExtendedAccessList(APIClassTemplate):
    """The ExtendedAccessList Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "entries"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/extendedaccesslists"
    REQUIRED_FOR_POST = ["name", "entries"]
    REQUIRED_FOR_PUT = ["id", "entries"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize ExtendedAccessList object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for ExtendedAccessList class.")
        self.parse_kwargs(**kwargs)


class ExtendedAccessListAce(object):
    """The ExtendedAccessListAce Object for creating ACEs or Lines in Extended ACLs"""

    VALID_JSON_DATA = [
        "action",
        "logging",
        "logLevel",
        "logInterval",
        "sourceNetworks",
        "sourcePorts",
        "destinationNetworks",
        "destinationPorts",
    ]

    def __init__(self, **kwargs) -> None:
        """
        Initialize ExtendedAccessListAce Object with default values.
        These defaults to permit any source/destination network and ports!!!

        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        self.action = "PERMIT"
        self.logging = "DEFAULT"
        self.logLevel = "INFORMATIONAL"
        self.logInterval = 300
        self.sourceNetworks = {}
        self.sourceNetworksObjects = []
        self.sourceNetworksLiterals = []
        self.sourcePorts = {}
        self.sourcePortsObjects = []
        self.sourcePortsLiterals = []
        self.destinationNetworks = {}
        self.destinationNetworksObjects = []
        self.destinationNetworksLiterals = []
        self.destinationPorts = {}
        self.destinationPortsObjects = []
        self.destinationPortsLiterals = []

    def build_ace(self, **kwargs):
        """
        Constructs dict that represents 1 ACE/Line in an Extended ACL.

        :param kwargs: Any other values passed during instantiation.
        :return: (dict) json_data
        """
        if self.sourceNetworksObjects:
            self.sourceNetworks["objects"] = self.sourceNetworksObjects

        if self.sourceNetworksLiterals:
            self.sourceNetworks["literals"] = self.sourceNetworksLiterals

        if self.sourcePortsObjects:
            self.sourcePorts["objects"] = self.sourcePortsObjects

        if self.sourcePortsLiterals:
            self.sourcePorts["literals"] = self.sourcePortsLiterals

        if self.destinationNetworksObjects:
            self.destinationNetworks["objects"] = self.destinationNetworksObjects

        if self.destinationNetworksLiterals:
            self.destinationNetworks["literals"] = self.destinationNetworksLiterals

        if self.destinationPortsObjects:
            self.destinationPorts["objects"] = self.destinationPortsObjects

        if self.destinationPortsLiterals:
            self.destinationPorts["literals"] = self.destinationPortsLiterals

        json_data = {}
        for key_value in self.__dict__:
            if key_value in self.VALID_JSON_DATA:
                if self.__dict__.get(key_value):
                    json_data[key_value] = self.__dict__.get(key_value)
        return json_data



"""PreFilter Rules Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.policy_services.prefilterpolicies import PreFilterPolicies
from fmcapi.api_objects.object_services.securityzones import SecurityZones
from fmcapi.api_objects.helper_functions import get_networkaddress_type
from fmcapi.api_objects.object_services.fqdns import FQDNS
from fmcapi.api_objects.object_services.networkgroups import NetworkGroups
from fmcapi.api_objects.object_services.networkaddresses import NetworkAddresses
from fmcapi.api_objects.object_services.protocolportobjects import ProtocolPortObjects
from fmcapi.api_objects.object_services.portobjectgroups import PortObjectGroups
from fmcapi.api_objects.object_services.vlangrouptags import VlanTags, VlanGroupTags
import logging


class PreFilterRules(APIClassTemplate):
    """The PreFilterRules object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "action",
        "sourceNetworks",
        "destinationNetworks",
        "sourceInterfaces",
        "destinationInterfaces",
        "sourcePorts",
        "destinationPorts",
        "vlanTags",
        "ruleType",
        "type",
        "enabled",
        "bidirectional",
        "encapsulationPorts",
        "enableSyslog",
        "logBegin",
        "logEnd",
        "sendEventsToFMC",
        "newComments",
        "commentHistoryList",
    ]
    PREFIX_URL = "/policy/prefilterpolicies"
    VALID_FOR_ACTION = ["FASTPATH", "ANALYZE", "BLOCK"]
    VALID_FOR_RULETYPE = ["TUNNEL", "PREFILTER"]
    VALID_FOR_ENCAPSULATIONPORTS = ["GRE", "IP_IN_IP", "IPV6_IN_IP", "TEREDO"]
    REQUIRED_FOR_POST = ["prefilter_id"]
    REQUIRED_FOR_GET = ["prefilter_id"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "prefilter_id",
        "prefilter_name",
        "insertBefore",
        "insertAfter",
        "section",
    ]

    def __init__(self, fmc, **kwargs):
        """
        Initialize PreFilterRules object.

        Set self.type to "PreFilterRule", parse the kwargs, and set up the self.URL.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for PreFilterRules class.")
        # self.type = "PreFilterRules"
        self.prefilter_id = None
        self.prefilter_added_to_url = False
        self.action = "FASTPATH"
        self.ruleType = "PREFILTER"
        self.bidirectional = False
        self.enableSyslog = False
        self.logBegin = False
        self.logEnd = False
        self.sendEventsToFMC = False
        self.enabled = False
        self.newComments = []
        self.commentHistoryList = []
        self.parse_kwargs(**kwargs)
        self.URL = f"{self.URL}{self.URL_SUFFIX}"
        self.type = "PrefilterRule"

    @property
    def URL_SUFFIX(self):
        """
        Add the URL suffixes for insertBefore and insertAfter.

        NOTE: You must specify these at the time of the object is initialized
        for this to work correctly. Example:
            This works:
                new_rule = PreFilterRules(fmc=fmc, prefilter_name='pre1', insertBefore=2)

            This does not:
                new_rule = PreFilterRules(fmc=fmc, prefilter_name='pre1')
                new_rule.insertBefore = 2
        """
        url = "?"
        if "insertBefore" in self.__dict__:
            url = f"{url}insertBefore={self.insertBefore}"
        elif "insertAfter" in self.__dict__:
            url = f"{url}insertAfter={self.insertAfter}"
        else:
            return ""
        return url

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and load into object properties.

        :param kwargs: (dict)
        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for PreFilterRules class")

        if "prefilter_id" in kwargs:
            self.prefilter(prefilter_id=kwargs["prefilter_id"])
        if "prefilter_name" in kwargs:
            self.prefilter(name=kwargs["prefilter_name"])
        if "action" in kwargs:
            self.validate_action(kwargs["action"])
        if "encapsulationPorts" in kwargs:
            self.validate_encapsulation_ports(kwargs["encapsulationPorts"])
        if "ruleType" in kwargs:
            self.rule_type(kwargs["ruleType"])
        if "enableSyslog" in kwargs:
            self.enableSyslog = kwargs["enableSyslog"]
        if "logBegin" in kwargs:
            self.logBegin = kwargs["logBegin"]
        if "logEnd" in kwargs:
            self.logEnd = kwargs["logEnd"]
        if "sendEventsToFMC" in kwargs:
            self.sendEventsToFMC = kwargs["sendEventsToFMC"]
        if "enabled" in kwargs:
            self.enabled = kwargs["enabled"]
        if "newComments" in kwargs:
            self.newComments = kwargs["newComments"]
        if "commentHistoryList" in kwargs:
            self.commentHistoryList = kwargs["commentHistoryList"]

    def prefilter(self, name=None, prefilter_id=None):
        """
        If the name has been supplied, go and pull the prefilter ID. If ID has been specified, add it to self.

        Also create the URL for prefilter policy
        :param name: (str) Prefilter name
        :param prefilter_id: (str) UUID of prefilter
        :return: None
        """
        logging.debug("In prefilter() for PreFilterRules")
        if not name and not prefilter_id:
            logging.warning(f"Unable to find prefilter as no name or ID was specified")

        elif name:
            logging.info(f'Searching ID for prefilter "{name}"')
            pre1 = PreFilterPolicies(fmc=self.fmc)
            pre1.get(name=name)

            if "id" in pre1.__dict__:
                prefilter_id = pre1.id
            else:
                logging.warning(
                    f"Prefilter policy {name} not found. Cannot setup prefilter policy for PreFilterRules"
                )

        self.prefilter_id = prefilter_id
        self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.prefilter_id}/prefilterrules"
        self.prefilter_added_to_url = True

    def validate_action(self, action):
        """
        Validate action against VALID_FOR_ACTION and sets property.

        :param action: (str): Prefilter rule action via VALID_FOR_ACTION constant.
        :return: None
        """
        if action in self.VALID_FOR_ACTION:
            self.action = action
        else:
            logging.warning(
                f"Action {action} is not a valid option\nValid actions are: {self.VALID_FOR_ACTION}"
            )

    def validate_encapsulation_ports(self, encapsulation_ports):
        """
        Validate encapsulation_ports against VALID_FOR_ENCAPSULATIONPORTS and sets property.

        :param encapsulation_ports: (list): Prefilter rule type via VALID_FOR_ENCAPSULATIONPORTS constant.
        :return: None
        """
        if all(
            port in self.VALID_FOR_ENCAPSULATIONPORTS for port in encapsulation_ports
        ):
            self.encapsulationPorts = encapsulation_ports
        else:
            logging.warning(
                f"Invalid protocol. Valid encapsulation ports are: {self.VALID_FOR_ENCAPSULATIONPORTS}"
            )

    def new_comments(self, action, value):
        """
        Add a comment to the comment list
        Args:
            action (str): Add, remove or clear
            value (str): Comment value to add
        """
        if action == "add":
            self.newComments.append(value)
        if action == "remove":
            self.newComments.remove(value)
        if action == "clear":
            self.newComments = []

    def source_interface(self, action, name=""):
        """
        Set the source interface (zone) information.

        :param action: (str) "add" or "remove"
        :param name: (str) Name of interface zone object
        :return: None
        """
        logging.debug("In source_zone() for PreFilterRules class.")
        if action == "clear":
            if "sourceInterfaces" in self.__dict__:
                del self.sourceInterfaces
                logging.info(f"Removed source zone from prefilter rule")

        elif action == "add":
            src_zone_id, src_zone_type = self.get_zone_id(name)
            if src_zone_id:
                self.add_zone("sourceInterfaces", name, src_zone_id, src_zone_type)
            else:
                logging.warning(f'Security zone object "{name}" not found on FMC')

    def destination_interface(self, action, name=""):
        """
        Set the destination interface (zone) information.

        :param action: (str) "add" or "remove"
        :param name: (str) Name of interface zone object
        :return: None
        """
        logging.debug("In source_zone() for PreFilterRules class.")
        if action == "clear":
            if "sourceZones" in self.__dict__:
                del self.destinationInterfaces
                logging.info(f"Removed source zone from prefilter rule")

        elif action == "add":
            dst_zone_id, dst_zone_type = self.get_zone_id(name)
            if dst_zone_id:
                self.add_zone("destinationInterfaces", name, dst_zone_id, dst_zone_type)
            else:
                logging.warning(f'Security zone object "{name}" not found on FMC')

    def get_zone_id(self, name):
        """
        Pull the ID for a security zone.

        :param name: (str) Name of interface zone object
        :return: (str) UUID of zone object
        """
        sec_zone = SecurityZones(fmc=self.fmc)
        sec_zone.get(name=name)
        if "id" in sec_zone.__dict__:
            return sec_zone.id, sec_zone.type
        else:
            return None, None

    def add_zone(self, target, name, id, zone_type):
        """
        Check if zone is already on object.

        Skip if it is, add if it isn't.  Create sourceZone if that object attribute doesn't exist at all.
        :param target: (str) "sourceZones" or "destinationZones"
        :param name: (str) Name of zone object
        :param id: (str) UUID of zone object
        :param zone_type: (str) Security zone type
        :return: None
        """
        if target in self.__dict__:
            # Look through existing zones, if present, for the zone name. Skip if it's already there
            zone_name_duplicate = False
            zone_list = []
            for zone in getattr(self, target)["objects"]:
                zone_list.append(zone)
                if zone["name"] == name:
                    zone_name_duplicate = True

            if not zone_name_duplicate:
                zone_list.append({"name": name, "id": id, "type": zone_type})
                setattr(self, target, {"objects": zone_list})

        # Set the zone if it doesn't exist
        else:
            setattr(
                self, target, {"objects": [{"name": name, "id": id, "type": zone_type}]}
            )

    def source_network(self, action, name=None, literal=None):
        """
        Add either an object or a literal to sourceNetworks.

        :param action: (str) Action to be done on this object
        :param name: (str) Object name
        :param literal: (str) Host, network or range
        :return: None
        """
        logging.debug("In source_network() for PreFilterRules class.")
        if literal and name:
            raise ValueError(
                "Adding source literal and object at the same time not supported"
            )
            return

        if not hasattr(self, "sourceNetworks"):
            self.sourceNetworks = {"objects": [], "literals": []}

        if action == "add" and literal:
            literal_type = get_networkaddress_type(literal)
            self.sourceNetworks["literals"].append(
                {"type": literal_type, "value": literal}
            )
            return

        if name:
            new_object = self.find_object(name)

        if not new_object:
            return

        if action == "add":
            # Check if object is already in the list and if not, then add it
            if new_object not in self.sourceNetworks["objects"]:
                logging.info(f'Adding "{name}" to sourceNetworks for prefilter rule')
                self.sourceNetworks["objects"].append(new_object)

        elif action == "remove":
            index = self.sourceNetworks["objects"].index(new_object)
            logging.info(f'Removing "{new_object}" from sourceNetworks')
            self.sourceNetworks["objects"].pop(index)

        elif action == "clear":
            logging.info("Clearing all destination networks")
            del self.destinationNetworks

    def destination_network(self, action, name=None, literal=None):
        """
        Add either an object or a literal to destinationNetworks.

        :param action: (str) Action to be done on this object
        :param name: (str) Object name
        :param literal: (str) Host, network or range
        :return: None
        """
        logging.debug("In destination_network() for PreFilterRules class.")
        if literal and name:
            raise ValueError(
                "Adding source literal and object at the same time not supported"
            )
            return

        if not hasattr(self, "destinationNetworks"):
            self.destinationNetworks = {"objects": [], "literals": []}

        if action == "add" and literal:
            literal_type = get_networkaddress_type(literal)
            self.destinationNetworks["literals"].append(
                {"type": literal_type, "value": literal}
            )
            return

        if name:
            new_object = self.find_object(name)

        if not new_object:
            return

        if action == "add":
            # Check if object is already in the list and if not, then add it
            if new_object not in self.destinationNetworks["objects"]:
                logging.info(
                    f'Adding "{name}" to destinationNetworks for prefilter rule'
                )
                self.destinationNetworks["objects"].append(new_object)

        elif action == "remove":
            index = self.destinationNetworks["objects"].index(new_object)
            logging.info(f'Removing "{new_object}" from destinationNetworks')
            self.destinationNetworks["objects"].pop(index)

        elif action == "clear":
            logging.info("Clearing all destination networks")
            del self.destinationNetworks

    def find_object(self, name):
        """
        Search through IP net objects, network group objects and FQDN objects by name.

        :param name: (str) Name of object
        :return: (dict) Object details or None
        """
        object_id, object_type = self.find_network_object(name)
        if object_id:
            return {"name": name, "id": object_id, "type": object_type}

        object_id, object_type = self.find_group_object(name)
        if object_id:
            return {"name": name, "id": object_id, "type": object_type}

        object_id, object_type = self.find_fqdn_object(name)
        if object_id:
            return {"name": name, "id": object_id, "type": object_type}

        logging.warning(f'Unable to find network object "{name}"')
        return None

    def find_network_object(self, name):
        """
        Search for network object by name.

        :param name: (str) Name of network object
        :return: id (str), type (str) or None None
        """
        network_object = NetworkAddresses(fmc=self.fmc, name=name)
        resp = network_object.get()

        return self._return_id_type(resp)

    def find_group_object(self, name):
        """
        Search for group object by name.

        :param name: (str) Name of group object
        :return: id (str), type (str) or None None
        """
        group_object = NetworkGroups(fmc=self.fmc, name=name)
        resp = group_object.get()

        return self._return_id_type(resp)

    def find_fqdn_object(self, name):
        """
        Search for group object by name.

        :param name: (str) Name of group object
        :return: id (str), type (str) or None None
        """
        fqdn_object = FQDNS(fmc=self.fmc, name=name)
        resp = fqdn_object.get()

        return self._return_id_type(resp)

    @staticmethod
    def _return_id_type(resp):
        """
        Check dict for ID and return ID and type if present.

        :param resp: (dict) Response from NetworkGroup, NetworkAddress or FQDNS GET
        :return: id (str), type (str) or None None
        """
        if "id" in resp.keys():
            return resp["id"], resp["type"]
        else:
            return None, None

    def rule_type(self, rule_type):
        """
        Set the rule type attribute.

        :param rule_type: (str) 'PREFILTER' or 'TUNNEL'
        :return: None
        """
        if rule_type in self.VALID_FOR_RULETYPE:
            self.ruleType = rule_type
        else:
            logging.warning(
                f'Rule type "{rule_type}" is not valid. Must be "PREFILTER" or "TUNNEL"'
            )

    def source_port(self, action, name=None, literal=None):
        """
        Create the source protocol and ports.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of port object
        :param literal: (dict) Dictionary of protocol and port expressed as integers
        """
        logging.debug("In source_port for PreFilterRules")

        if literal and name:
            raise ValueError(
                "Adding source literal and object at the same time not supported"
            )
            return

        if not hasattr(self, "sourcePorts"):
            self.sourcePorts = {"literals": [], "objects": []}

        if action == "add" and literal:
            if self._port_literal_verify(literal):
                literal["type"] = "PortLiteral"
                self.sourcePorts["literals"].append(literal)
            return

        if name:
            port_object = self._find_port_object(name)

        if not port_object:
            return

        if action == "add":
            if port_object not in self.sourcePorts["objects"]:
                logging.info(f'Adding "{port_object}" to sourcePorts')
                self.sourcePorts["objects"].append(port_object)

        elif action == "remove":
            index = self.sourcePorts["objects"].index(port_object)
            logging.info(f'Removing "{port_object}" from sourcePorts')
            self.sourcePorts["objects"].pop(index)

        elif action == "clear":
            del self.sourcePorts

    def destination_port(self, action, name=None, literal=None):
        """
        Create the destination protocol and ports.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of port object
        :param literal: (dict) Dictionary of protocol and port expressed as integers
        """
        logging.debug("In destination_port for PreFilterRules")

        if literal and name:
            raise ValueError(
                "Adding destination literal and object at the same time not supported"
            )
            return

        if not hasattr(self, "destinationPorts"):
            self.destinationPorts = {"literals": [], "objects": []}

        if action == "add" and literal:
            if self._port_literal_verify(literal):
                literal["type"] = "PortLiteral"
                self.destinationPorts["literals"].append(literal)
            return

        if name:
            port_object = self._find_port_object(name)

        if not port_object:
            return

        if action == "add":
            if port_object not in self.destinationPorts["objects"]:
                logging.info(f'Adding "{port_object}" to destinationPorts')
                self.destinationPorts["objects"].append(port_object)

        elif action == "remove":
            index = self.destinationPorts["objects"].index(port_object)
            logging.info(f'Removing "{port_object}" from destinationPorts')
            self.destinationPorts["objects"].pop(index)

        elif action == "clear":
            del self.destinationPorts

    def _find_port_object(self, name):
        """
        Find port object or port group object and return dictionary.

        :param name: (str) Name of port object/port group object
        :return: (dict) Port object
        """
        protocol_port = ProtocolPortObjects(fmc=self.fmc, name=name)
        resp = protocol_port.get()
        if "id" in resp.keys():
            return {"name": name, "id": resp["id"], "type": resp["type"]}

        protocol_port_group = PortObjectGroups(fmc=self.fmc, name=name)
        resp = protocol_port_group.get()
        if "id" in resp.keys():
            return {"name": name, "id": resp["id"], "type": resp["type"]}

        logging.warning(f'Unable to find port object "{name}"')
        return None

    @staticmethod
    def _port_literal_verify(literal):
        """
        Ensure that the literal is expressed as integers.

        It's too hard dealing with named protocols/ports.  Check that protocol and ports keys are in the dictionary.

        :param literal: (dict) Dictionary with protocol and ports
        :return: (bool)
        """
        if not isinstance(literal, dict):
            logging.warning(
                'Invalid port literal. Must be defined as a dictionary: "{"protocol": <protocol>, "port": <port}'
            )
            return False

        if "protocol" not in literal.keys() and "port" not in literal.keys():
            logging.warning(
                '"protocol" and/or "port" missing from literal. You must specify both'
            )
            return False

        return True

    def encapsulation_port(self, action, value=None):
        """
        Set the encapsulation port for the tunnel rule.

        :param action: (str) "add", "remove" or "clear"
        :param value: (str) Name of encapsulation port
        :return: None
        """
        logging.debug("In encapsulation_port() for PreFilterRules class.")
        if not hasattr(self, "encapsulationPorts"):
            self.encapsulationPorts = []

        if action == "add":
            self.encapsulationPorts.append(value)

        elif action == "remove":
            self.encapsulationPorts.remove(value)

        elif action == "clear":
            del self.encapsulationPorts
            logging.info(f"Removed encapsulation ports from the prefilter rule")

    def vlan_tags(self, action, name=None, literal=None):
        """
        Add, remove or clear VLAN tags.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of VLAN tag object
        :param literal: (str) VLAN tag or range
        :return: None
        """
        logging.debug("In vlan_tags() for PreFilterRules")
        if literal and name:
            raise ValueError(
                "Adding/Removing VLAN literal and object at the same time not supported"
            )
            return

        if not hasattr(self, "vlanTags"):
            self.vlanTags = {"literals": [], "objects": []}

        if action == "add" and literal:
            start_vlan, end_vlan = self._vlan_tag_check(literal)
            self.vlanTags["literals"].append(
                {"startTag": start_vlan, "endTag": end_vlan, "type": "VlanTagLiteral"}
            )
            return

        if name:
            vlan_object = self._find_vlan_object(name)

        if not vlan_object:
            return

        if action == "add":
            if vlan_object not in self.vlanTags["objects"]:
                self.vlanTags["objects"].append(vlan_object)

        if action == "remove":
            index = self.vlanTags["objects"].index(vlan_object)
            logging.info(f'Removing "{vlan_object}" from vlanTags')
            self.vlanTags["objects"].pop(index)

        elif action == "clear":
            del self.vlanTags

    @staticmethod
    def _vlan_tag_check(literal):
        """
        Check if the VLAN tag literal is a range or not and returns start and end tag number.

        :param literal: (str) VLAN tag literal
        :return: start_vlan (str), end_vlan (str)
        """
        if "-" in literal:
            vlans = literal.strip("-")
            return vlans[0], vlans[1]
        else:
            return literal, literal

    def _find_vlan_object(self, name):
        """
        Find the vlan or vlan group object by name.

        :param name: (str) Object name
        """
        vlan_object = VlanTags(fmc=self.fmc, name=name)
        resp = vlan_object.get()
        if "id" in resp.keys():
            return {"name": name, "id": resp["id"], "type": resp["type"]}

        vlan_object = VlanGroupTags(fmc=self.fmc, name=name)
        resp = vlan_object.get()
        if "id" in resp.keys():
            return {"name": name, "id": resp["id"], "type": resp["type"]}

        logging.warning(f'Unable to find vlan object "{name}"')
        return None



import logging
import fmcapi


def test__terminateravpnsessions(fmc):
    logging.info("Test TerminateRAVPNSessions. Requires an existing session id, user, or user+device id")
    logging.info("Terminating many sessions at once, like all sessions on a device, will likely require increasing fmc timeout")

    obj1 = fmcapi.TerminateRAVPNSessions(fmc=fmc)
    ######DIFFERENT TERMINATION METHODS##########
    # Terminate user/users on a specific deivce
    usernames_to_terminate = ['zlantztest']
    device_id = "5fceadc2-ffef-11ed-94e9-8f60148d3f9b"
    obj1.terminateBy = "USER"
    obj1.usernames = usernames_to_terminate
    obj1.deviceId = device_id

    # Terminate by specific session id
    # Currently have not found a reliable way of getting ravpnsession ids since there is no get methods for ravpnsessions
    # I have already complained about this to the BU
    # obj1.terminateBy = "SESSION"

    # Terminate all sessions on device
    # device_id = "5fceadc2-ffef-11ed-94e9-8f60148d3f9b"
    # obj1.terminateBy = "DEVICE"
    # obj1.deviceId = device_id

    result = obj1.post()
    logging.info(result)
    del(obj1)

    logging.info("Test TerminateRAVPNSessions done.\n")
